"""Tests asdf implementations of core module."""
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from asdf import ValidationError
from scipy.spatial.transform import Rotation

import weldx.transformations as tf
from weldx.asdf.utils import _write_buffer, _write_read_buffer
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.core import MathematicalExpression as ME  # nopep8
from weldx.core import TimeSeries
from weldx.transformations import WXRotation

# WXRotation ---------------------------------------------------------------------
_base_rotation = Rotation.from_euler(
    seq="xyz", angles=[[10, 20, 60], [25, 50, 175]], degrees=True
)


@pytest.mark.parametrize(
    "inputs",
    [
        _base_rotation,
        WXRotation.from_quat(_base_rotation.as_quat()),
        WXRotation.from_matrix(_base_rotation.as_matrix()),
        WXRotation.from_rotvec(_base_rotation.as_rotvec()),
        WXRotation.from_euler(seq="xyz", angles=[10, 20, 60], degrees=True),
        WXRotation.from_euler(seq="xyz", angles=[0.2, 1.3, 3.14], degrees=False),
        WXRotation.from_euler(seq="XYZ", angles=[10, 20, 60], degrees=True),
        WXRotation.from_euler(seq="y", angles=[10, 60, 40, 90], degrees=True),
        WXRotation.from_euler(seq="Z", angles=[10, 60, 40, 90], degrees=True),
        WXRotation.from_euler(
            seq="xy", angles=[[10, 10], [60, 60], [40, 40], [70, 75]], degrees=True
        ),
    ],
)
def test_rotation(inputs):
    data = _write_read_buffer({"rot": inputs})
    assert np.allclose(data["rot"].as_quat(), inputs.as_quat())


def test_rotation_euler_exception():
    with pytest.raises(ValueError):
        WXRotation.from_euler(seq="XyZ", angles=[10, 20, 60], degrees=True)


# xarray.DataArray ---------------------------------------------------------------------
def get_xarray_example_data_array():
    """
    Get an xarray.DataArray for test purposes.

    Returns
    -------
    xarray.DataArray
        DataArray for test purposes

    """
    data = np.array([[0, 1], [2, 3]])

    time_labels = ["2020-05-01", "2020-05-03"]
    d1 = np.array([-1, 1])
    d2 = pd.DatetimeIndex(time_labels)
    coords = {"d1": d1, "d2": d2, "time_labels": (["d2"], time_labels)}

    dax = xr.DataArray(data=data, dims=["d1", "d2"], coords=coords)

    dax.attrs = {"answer": 42}

    return dax


@pytest.mark.parametrize("copy_arrays", [True, False])
@pytest.mark.parametrize("lazy_load", [True, False])
def test_xarray_data_array(copy_arrays, lazy_load):
    """Test ASDF read/write of xarray.DataArray."""
    dax = get_xarray_example_data_array()
    tree = {"dax": dax}
    dax_file = _write_read_buffer(
        tree, open_kwargs={"copy_arrays": copy_arrays, "lazy_load": lazy_load}
    )["dax"]
    assert dax.identical(dax_file)


# xarray.Dataset ---------------------------------------------------------------------
def get_xarray_example_dataset():
    """
    Get an xarray.Dataset for test purposes.

    Returns
    -------
        Dataset for test purposes
    """

    temp_data = [
        [[15.0, 16.0, 17.0], [18.0, 19.0, 20.0]],
        [[21.0, 22.0, 23.0], [24.0, 25.0, 26.0]],
    ]
    temp = Q_(temp_data, "°C")
    precipitation = [
        [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
        [[7.0, 8.0, 9.0], [10.0, 11.0, 12.0]],
    ]
    lon = [[-99.83, -99.32], [-99.79, -99.23]]
    lat = [[42.25, 42.21], [42.63, 42.59]]

    dsx = xr.Dataset(
        {
            "temperature": (["x", "y", "time"], temp),
            "precipitation": (["x", "y", "time"], precipitation),
        },
        coords={
            "lon": (["x", "y"], lon),
            "lat": (["x", "y"], lat),
            "time": pd.date_range("2014-09-06", periods=3),
            "time_labels": (["time"], ["2014-09-06", "2014-09-09", "2014-09-12"]),
            "time_ref": pd.Timestamp("2014-09-05"),
        },
    )
    dsx.attrs = {"temperature": "Celsius"}
    return dsx


@pytest.mark.parametrize("copy_arrays", [True, False])
@pytest.mark.parametrize("lazy_load", [True, False])
def test_xarray_dataset(copy_arrays, lazy_load):
    dsx = get_xarray_example_dataset()
    tree = {"dsx": dsx}
    dsx_file = _write_read_buffer(
        tree, open_kwargs={"copy_arrays": copy_arrays, "lazy_load": lazy_load}
    )["dsx"]
    assert dsx.identical(dsx_file)


# weldx.transformations.LocalCoordinateSystem ------------------------------------------
def get_local_coordinate_system(time_dep_orientation: bool, time_dep_coordinates: bool):
    """
    Get a local coordinate system.

    Parameters
    ----------
    time_dep_orientation :
        If True, the coordinate system has a time dependent orientation.
    time_dep_coordinates :
        If True, the coordinate system has a time dependent coordinates.

    Returns
    -------
    weldx.transformations.LocalCoordinateSystem:
        A local coordinate system

    """
    if not time_dep_coordinates:
        coords = Q_(np.asarray([2.0, 5.0, 1.0]), "mm")
    else:
        coords = Q_(
            np.asarray(
                [[2.0, 5.0, 1.0], [1.0, -4.0, 1.2], [0.3, 4.4, 4.2], [1.1, 2.3, 0.2]]
            ),
            "mm",
        )

    if not time_dep_orientation:
        orientation = tf.rotation_matrix_z(np.pi / 3)
    else:
        orientation = tf.rotation_matrix_z(np.pi / 2 * np.array([1, 2, 3, 4]))

    if not time_dep_orientation and not time_dep_coordinates:
        return tf.LocalCoordinateSystem(orientation=orientation, coordinates=coords)

    time = pd.DatetimeIndex(["2000-01-01", "2000-01-02", "2000-01-03", "2000-01-04"])
    return tf.LocalCoordinateSystem(
        orientation=orientation, coordinates=coords, time=time
    )


@pytest.mark.parametrize("time_dep_orientation", [False, True])
@pytest.mark.parametrize("time_dep_coordinates", [False, True])
@pytest.mark.parametrize("copy_arrays", [True, False])
@pytest.mark.parametrize("lazy_load", [True, False])
def test_local_coordinate_system(
    time_dep_orientation, time_dep_coordinates, copy_arrays, lazy_load
):
    """Test (de)serialization of LocalCoordinateSystem in ASDF."""
    lcs = get_local_coordinate_system(time_dep_orientation, time_dep_coordinates)
    data = _write_read_buffer(
        {"lcs": lcs}, open_kwargs={"copy_arrays": copy_arrays, "lazy_load": lazy_load}
    )
    assert data["lcs"] == lcs


def test_local_coordinate_system_shape_violation():
    """Test if the shape validators work as expected."""
    # coordinates have wrong shape ------------------------
    orientation = xr.DataArray(
        data=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        dims=["u", "v"],
        coords={"u": ["x", "y", "z"], "v": [0, 1, 2]},
    )
    coordinates = xr.DataArray(data=[1, 2], dims=["c"], coords={"c": ["x", "y"]},)
    lcs = tf.LocalCoordinateSystem(
        orientation=orientation, coordinates=coordinates, construction_checks=False
    )

    with pytest.raises(ValidationError):
        _write_buffer({"lcs": lcs})

    # orientations have wrong shape -----------------------
    orientation = xr.DataArray(
        data=[[1, 2], [3, 4]], dims=["c", "v"], coords={"c": ["x", "y"], "v": [0, 1]},
    )
    coordinates = xr.DataArray(
        data=[1, 2, 3], dims=["u"], coords={"u": ["x", "y", "z"]},
    )
    lcs = tf.LocalCoordinateSystem(
        orientation=orientation, coordinates=coordinates, construction_checks=False
    )

    with pytest.raises(ValidationError):
        _write_buffer({"lcs": lcs})


# weldx.transformations.CoordinateSystemManager ----------------------------------------


def get_example_coordinate_system_manager():
    """Get a consistent CoordinateSystemManager instance for test purposes."""
    csm = tf.CoordinateSystemManager("root")
    csm.create_cs("lcs_01", "root", coordinates=[1, 2, 3])
    csm.create_cs(
        "lcs_02",
        "root",
        orientation=tf.rotation_matrix_z(np.pi / 3),
        coordinates=[4, -7, 8],
    )
    csm.create_cs(
        "lcs_03",
        "lcs_02",
        orientation=tf.rotation_matrix_y(np.pi / 11),
        coordinates=[4, -7, 8],
    )
    return csm


@pytest.mark.parametrize("copy_arrays", [True, False])
@pytest.mark.parametrize("lazy_load", [True, False])
def test_coordinate_system_manager(copy_arrays, lazy_load):
    csm = get_example_coordinate_system_manager()
    tree = {"cs_hierarchy": csm}
    data = _write_read_buffer(
        tree, open_kwargs={"copy_arrays": copy_arrays, "lazy_load": lazy_load}
    )
    csm_file = data["cs_hierarchy"]
    assert csm == csm_file


def get_coordinate_system_manager_with_subsystems(nested: bool):
    lcs = [tf.LocalCoordinateSystem(coordinates=[i, -i, -i]) for i in range(12)]

    # global system
    csm_global = tf.CoordinateSystemManager("base", "Global System", "2000-06-08")
    csm_global.add_cs("robot", "base", lcs[0])
    csm_global.add_cs("specimen", "base", lcs[1])

    # robot system
    csm_robot = tf.CoordinateSystemManager("robot", "Robot system")
    csm_robot.add_cs("head", "robot", lcs[2])

    # robot head system
    csm_head = tf.CoordinateSystemManager("head", "Head system")
    csm_head.add_cs("torch tcp", "head", lcs[3])
    csm_head.add_cs("camera tcp", "head", lcs[4], lsc_child_in_parent=False)
    csm_head.add_cs("scanner 1 tcp", "head", lcs[5])
    csm_head.add_cs("scanner 2 tcp", "head", lcs[6])

    # scanner system 1
    csm_scanner_1 = tf.CoordinateSystemManager("scanner 1", "Scanner 1 system")
    csm_scanner_1.add_cs("scanner 1 tcp", "scanner 1", lcs[7])

    # scanner system 2
    csm_scanner_2 = tf.CoordinateSystemManager("scanner 2", "Scanner 2 system")
    csm_scanner_2.add_cs("scanner 2 tcp", "scanner 2", lcs[8])

    # specimen system
    csm_specimen = tf.CoordinateSystemManager("specimen", "Specimen system")
    csm_specimen.add_cs("thermocouple 1", "specimen", lcs[9])
    csm_specimen.add_cs("thermocouple 2", "specimen", lcs[10])
    csm_specimen.add_cs("thermocouple 3", "thermocouple 2", lcs[11])

    if nested:
        csm_head.merge(csm_scanner_1)
        csm_head.merge(csm_scanner_2)
        csm_robot.merge(csm_head)
        csm_global.merge(csm_robot)
        csm_global.merge(csm_specimen)
    else:
        csm_global.merge(csm_specimen)
        csm_global.merge(csm_robot)
        csm_global.merge(csm_head)
        csm_global.merge(csm_scanner_1)
        csm_global.merge(csm_scanner_2)

    return csm_global


@pytest.mark.parametrize("copy_arrays", [True, False])
@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("nested", [True, False])
def test_coordinate_system_manager_with_subsystems(copy_arrays, lazy_load, nested):
    csm = get_coordinate_system_manager_with_subsystems(nested)
    tree = {"cs_hierarchy": csm}
    data = _write_read_buffer(
        tree, open_kwargs={"copy_arrays": copy_arrays, "lazy_load": lazy_load}
    )
    csm_file = data["cs_hierarchy"]
    assert csm == csm_file


@pytest.mark.parametrize("copy_arrays", [True, False])
@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("csm_time_ref", [None, "2000-03-16"])
def test_coordinate_system_manager_time_dependencies(
    copy_arrays, lazy_load, csm_time_ref
):
    """Test serialization of time components from CSM and its attached LCS."""
    lcs_tdp_1_time_ref = None
    if csm_time_ref is None:
        lcs_tdp_1_time_ref = pd.Timestamp("2000-03-17")
    lcs_tdp_1 = tf.LocalCoordinateSystem(
        coordinates=[[1, 2, 3], [4, 5, 6]],
        time=pd.TimedeltaIndex([1, 2], "D"),
        time_ref=lcs_tdp_1_time_ref,
    )
    lcs_tdp_2 = tf.LocalCoordinateSystem(
        coordinates=[[3, 7, 3], [9, 5, 8]],
        time=pd.TimedeltaIndex([1, 2], "D"),
        time_ref=pd.Timestamp("2000-03-21"),
    )

    csm_root = tf.CoordinateSystemManager("root", "csm_root", csm_time_ref)
    csm_root.add_cs("cs_1", "root", lcs_tdp_2)

    csm_sub_1 = tf.CoordinateSystemManager("cs_2", "csm_sub_1", csm_time_ref)
    csm_sub_1.add_cs("cs_1", "cs_2", lcs_tdp_2)
    csm_sub_1.add_cs("cs_3", "cs_2", lcs_tdp_1)

    csm_sub_2 = tf.CoordinateSystemManager("cs_4", "csm_sub_2")
    csm_sub_2.create_cs("cs_1", "cs_4")

    csm_root.merge(csm_sub_1)
    csm_root.merge(csm_sub_2)

    tree = {"cs_hierarchy": csm_root}
    data = _write_read_buffer(
        tree, open_kwargs={"copy_arrays": copy_arrays, "lazy_load": lazy_load}
    )
    csm_file = data["cs_hierarchy"]
    assert csm_root == csm_file


# --------------------------------------------------------------------------------------
# TimeSeries
# --------------------------------------------------------------------------------------


@pytest.mark.parametrize("copy_arrays", [True, False])
@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize(
    "ts",
    [
        TimeSeries(Q_(42, "m")),
        TimeSeries(Q_([42, 23, 12], "m"), time=pd.TimedeltaIndex([0, 2, 4])),
        TimeSeries(Q_([42, 23, 12], "m"), time=pd.TimedeltaIndex([0, 2, 5])),
        TimeSeries(ME("a*t+b", parameters={"a": Q_(2, "1/s"), "b": Q_(5, "")})),
    ],
)
def test_time_series_discrete(ts, copy_arrays, lazy_load):
    ts_file = _write_read_buffer(
        {"ts": ts}, open_kwargs={"copy_arrays": copy_arrays, "lazy_load": lazy_load}
    )["ts"]
    if isinstance(ts.data, ME):
        assert ts.data == ts_file.data
    else:
        assert np.all(ts_file.data == ts.data)
    assert np.all(ts_file.time == ts.time)
    assert ts_file.interpolation == ts.interpolation



"""Test functions of the visualization package."""

import matplotlib.pyplot as plt
import pandas as pd
import pytest

# pylint: disable=W0611
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

import weldx.transformations as tf
import weldx.visualization as vs

# pylint: enable=W0611


def test_plot_coordinate_system():
    """Test executing all possible code paths."""
    lcs_constant = tf.LocalCoordinateSystem()

    time = pd.TimedeltaIndex([10, 11, 12], "s")
    orientation_tdp = [
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        [[0, 1, 0], [-1, 0, 0], [0, 0, 1]],
        [[-1, 0, 0], [0, -1, 0], [0, 0, 1]],
    ]
    coordinates_tdp = [[0, 0, 1], [0, 0, 2], [0, -1, 0]]
    lcs_tdp = tf.LocalCoordinateSystem(
        orientation=orientation_tdp, coordinates=coordinates_tdp, time=time
    )

    fig = plt.figure()
    ax = fig.gca(projection="3d")

    vs.plot_coordinate_system(lcs_constant, ax, "g")
    vs.plot_coordinate_system(lcs_tdp, ax, "r", "2016-01-10")
    vs.plot_coordinate_system(lcs_tdp, ax, "b", "2016-01-11", time_idx=1)
    vs.plot_coordinate_system(
        lcs_tdp, ax, "y", "2016-01-12", pd.TimedeltaIndex([12], "s")
    )

    # exceptions ------------------------------------------

    # label without color
    with pytest.raises(Exception):
        vs.plot_coordinate_system(lcs_constant, ax, label="label")


def test_set_axes_equal():
    """Test executing all possible code paths."""
    fig = plt.figure()
    ax = fig.gca(projection="3d")
    vs.set_axes_equal(ax)



"""Test ASDF serialization of AWS schema definitions."""
import pytest

# weld design -----------------------------------------------------------------
from weldx.asdf.tags.weldx.aws.design.base_metal import BaseMetal
from weldx.asdf.tags.weldx.aws.design.connection import Connection

# weld design -----------------------------------------------------------------
from weldx.asdf.tags.weldx.aws.design.joint_penetration import JointPenetration
from weldx.asdf.tags.weldx.aws.design.sub_assembly import SubAssembly
from weldx.asdf.tags.weldx.aws.design.weld_details import WeldDetails
from weldx.asdf.tags.weldx.aws.design.weldment import Weldment
from weldx.asdf.tags.weldx.aws.design.workpiece import Workpiece

# welding process -----------------------------------------------------------------
from weldx.asdf.tags.weldx.aws.process.arc_welding_process import ArcWeldingProcess
from weldx.asdf.tags.weldx.aws.process.gas_component import GasComponent
from weldx.asdf.tags.weldx.aws.process.shielding_gas_for_procedure import (
    ShieldingGasForProcedure,
)
from weldx.asdf.tags.weldx.aws.process.shielding_gas_type import ShieldingGasType

# iso groove -----------------------------------------------------------------
from weldx.asdf.tags.weldx.core.iso_groove import get_groove
from weldx.asdf.utils import _write_read_buffer
from weldx.constants import WELDX_QUANTITY as Q_


def test_aws_example():
    """Test validity of current AWS Data Dictionary standard implementation."""
    # welding process -----------------------------------------------------------------
    gas_comp = [
        GasComponent("argon", Q_(82, "percent")),
        GasComponent("carbon dioxide", Q_(18, "percent")),
    ]
    gas_type = ShieldingGasType(gas_component=gas_comp, common_name="SG")

    gas_for_procedure = ShieldingGasForProcedure(
        use_torch_shielding_gas=True,
        torch_shielding_gas=gas_type,
        torch_shielding_gas_flowrate=Q_(20, "l / min"),
    )

    arc_welding_process = ArcWeldingProcess("GMAW")
    with pytest.raises(ValueError):  # test for non viable process string
        ArcWeldingProcess("NON_EXISTENT_PROCESS")

    process = {
        "arc_welding_process": arc_welding_process,
        "shielding_gas": gas_for_procedure,
    }

    # weld design -----------------------------------------------------------------
    v_groove = get_groove(
        groove_type="VGroove",
        workpiece_thickness=Q_(9, "mm"),
        groove_angle=Q_(50, "deg"),
        root_face=Q_(4, "mm"),
        root_gap=Q_(2, "mm"),
    )
    u_groove = get_groove(
        groove_type="UGroove",
        workpiece_thickness=Q_(15, "mm"),
        bevel_angle=Q_(9, "deg"),
        bevel_radius=Q_(6, "mm"),
        root_face=Q_(3, "mm"),
        root_gap=Q_(1, "mm"),
    )

    joint_penetration = JointPenetration(
        complete_or_partial="completePenetration", root_penetration=Q_(1.0, "mm")
    )
    weld_details = WeldDetails(
        joint_design=v_groove, weld_sizes=Q_(320, "mm"), number_of_passes=1
    )
    weld_details2 = WeldDetails(
        joint_design=u_groove, weld_sizes=Q_(320, "mm"), number_of_passes=1
    )
    connection1 = Connection(
        joint_type="butt_joint",
        weld_type="singleVGroove",
        joint_penetration=joint_penetration,
        weld_details=weld_details,
    )
    connection2 = Connection(
        joint_type="butt_joint",
        weld_type="singleUGroove",
        joint_penetration=joint_penetration,
        weld_details=weld_details2,
    )
    workpieces = [Workpiece(geometry="V-Groove")]
    sub_assembly = [
        SubAssembly(workpiece=workpieces, connection=connection1),
        SubAssembly(workpiece=workpieces, connection=connection2),
    ]

    weldment = Weldment(sub_assembly)

    base_metal = BaseMetal("steel", "plate", Q_(10.3, "mm"))

    tree = dict(process=process, weldment=weldment, base_metal=base_metal)

    data = _write_read_buffer(tree)
    assert isinstance(data, dict)



"""Generate example implementation of AWS "A Welding Data Dictionary."""

from weldx.asdf.utils import create_asdf_dataclass

# design -----------------------------------------------------------------

"""
create_asdf_dataclass(
    asdf_name="aws/NAME",
    asdf_version="1.0.0",
    class_name="PythonClass",
    schema_description=""DESCRIPTION",
    properties=[
        "property1",
        "property2",
        "property3",
        "property4",
        "property5",
        "property6",
        "property7",
        "property8",
    ],
    required=[
        "property1",
        "property2",
        "property3",
        "property4",
        "property5",
        "property6",
        "property7",
        "property8",
    ],
    property_order=[
        "property1",
        "property2",
        "property3",
        "property4",
        "property5",
        "property6",
        "property7",
        "property8",
    ],
    property_types=[
        "property1",
        "property2",
        "property3",
        "property4",
        "property5",
        "property6",
        "property7",
        "property8",
    ],
    description=[
        "property1",
        "property2",
        "property3",
        "property4",
        "property5",
        "property6",
        "property7",
        "property8",
    ],
)
"""

# design -----------------------------------------------------------------

# design/weldment -----------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/design/weldment",
    asdf_version="1.0.0",
    class_name="Weldment",
    schema_description="An assembly whose component parts are joined by welding",
    properties=["sub_assembly"],
    required=["sub_assembly"],
    property_order=["sub_assembly"],
    property_types=["List[SubAssembly]"],
    description=["A list of assemblies composing the weldment"],
)

# design/sub_assembly -----------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/design/sub_assembly",
    asdf_version="1.0.0",
    class_name="SubAssembly",
    schema_description="Two workpieces joined by a weld. Identifies simple weldments that can be joined to form a largermore complicated weldment.",
    properties=["workpiece", "connection"],
    required=["workpiece", "connection"],
    property_order=["workpiece", "connection"],
    property_types=["List[Workpiece]", "Connection"],
    description=[
        "Describes a piece of metal, its materials and geometries.",
        "The joint formed by edge preparation of two workpieces and the weld that joins the workpieces together.",
    ],
)

# design/workpiece -----------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/design/workpiece",
    asdf_version="1.0.0",
    class_name="Workpiece",
    schema_description="""This element describes a piece of metal, its material, geometric properties and boundaries, including its edge shape for the welded joint. A workpiece can be a single piece of plate, or a previously assembled weldment. The scope of this data is beyond this welding data dictionary. This element represents an interface with CAD applications.""",
    properties=["geometry"],
    required=["geometry"],
    property_order=["geometry"],
    property_types=["str"],
    description=["The workpiece geometry."],
)


# design/connection -----------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/design/connection",
    asdf_version="1.0.0",
    class_name="Connection",
    schema_description="The joint formed by edge preparation of two workpieces and the weld that joins the workpieces together.",
    properties=["joint_type", "weld_type", "joint_penetration", "weld_details"],
    required=["joint_type", "weld_type", "joint_penetration", "weld_details"],
    property_order=["joint_type", "weld_type", "joint_penetration", "weld_details"],
    property_types=["str", "str", "JointPenetration", "WeldDetails"],
    description=[
        "A weld joint classification based the relative orientation of the members being joined. TODO ENUM",
        "A region of coalescence of materials produced by heating or pressure, that joins two pieces of metal. TODO ENUM",
        "joint_penetration",
        "weld_details",
    ],
)

# design/joint_penetration ------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/design/joint_penetration",
    asdf_version="1.0.0",
    class_name="JointPenetration",
    schema_description="Various dimensions of penetration of the weld into the base materials",
    properties=[
        "complete_or_partial",
        "root_penetration",
        "groove_weld_size",
        "incomplete_joint_penetration",
        "weld_size",
        "weld_size_E1",
        "weld_size_E2",
        "depth_of_fusion",
    ],
    required=["complete_or_partial", "units", "root_penetration"],
    property_order=[
        "complete_or_partial",
        "root_penetration",
        "groove_weld_size",
        "incomplete_joint_penetration",
        "weld_size",
        "weld_size_E1",
        "weld_size_E2",
        "depth_of_fusion",
    ],
    property_types=[
        "str",
        "pint.Quantity",
        "float",
        "float",
        "float",
        "float",
        "float",
        "float",
    ],
    description=[
        "The weld design calls for partial or complete penetration. TODO ENUM",
        "The distance the weld metal extends into the root joint",
        "The joint penetration of a groove weld",
        'See AWS A3.0 definition, and Figure 26 - "Joint Penetration...", Figure 30 - Fusion Welds',
        'See AWS A3.0 definition, and Figure 26 - "Joint Penetration...", Figure 30 - Fusion Welds',
        'See AWS A3.0 definition, and Figure 26 - "Joint Penetration...", Figure 30 - Fusion Welds',
        'See AWS A3.0 definition, and Figure 26 - "Joint Penetration...", Figure 30 - Fusion Welds',
        "The distance that fusion extends into the base metal or the previous bead from the surface melted during welding",
    ],
)

# design/weld_details -----------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/design/weld_details",
    asdf_version="1.0.0",
    class_name="WeldDetails",
    schema_description="Information describing the joint and the geometric parameters of the designed weld.",
    properties=["joint_design", "weld_sizes", "number_of_passes"],
    required=["joint_design", "weld_sizes", "number_of_passes"],
    property_order=["joint_design", "weld_sizes", "number_of_passes"],
    property_types=["str", "pint.Quantity", "int"],
    description=[
        'AWS 3.0, p. 19 : "Detailed Description of mating configuration of two components, sufficient information for shop preparation of the components." TODO CLASS',
        "Linear dimensions of cross sectional weld features. Sizes are of design and actual features. TODO CLASS",
        "Number of times needed to deposit weld metal to complete the weld.",
    ],
)

# design/base_metal -----------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/design/base_metal",
    asdf_version="1.0.0",
    class_name="BaseMetal",
    schema_description="The material that is being joined using a weld.",
    properties=[
        "common_name",
        "m_number",
        "group_number",
        "product_form",
        "thickness",
        "diameter",
        "specification_number",
        "specification_version",
        "specification_organization",
        "UNS_number",
        "CAS_number",
        "heat_lot_identification",
        "composition",
        "manufacturing_history",
        "service_history",
        "applied_coating_specification",
    ],
    required=["common_name", "product_form", "thickness"],
    property_order=[
        "common_name",
        "m_number",
        "group_number",
        "product_form",
        "thickness",
        "diameter",
        "specification_number",
        "specification_version",
        "specification_organization",
        "UNS_number",
        "CAS_number",
        "heat_lot_identification",
        "composition",
        "manufacturing_history",
        "service_history",
        "applied_coating_specification",
    ],
    property_types=[
        "str",
        "str",
        "str",
        "str",
        "float",
        "float",
        "str",
        "str",
        "str",
        "str",
        "str",
        "str",
        "str",
        "str",
        "str",
        "str",
    ],
    description=[
        "The trade name, a name used without all the designations of the formal specification.",
        "A designation used to group base metals for procedure and performance qualifications.",
        "A classification system for metal by material properties.",
        "The form of the workpieces to be joined. TODO: ENUM",
        "Plate or sheet thickness, if tube, specifies wall thickness.",
        "Outside diameter, only used if material is tube.",
        "The standard designation of the formal material classification.",
        "Version of the specification used.",
        "The organization responsible for generating the specification.",
        "Unified Numbering System for Metals and Alloys, managed by ASTM and SAE.",
        "Chemical Abstracts Service Registry Number, a unique identifier for substances issued by the Chemical Abstracts Service.",
        "A unique identifier issued by a materials manufacturer assigned to manufacturing batches.",
        "Detailed chemical composition, by elements. This Type needs expansion.",
        "Mechanical manufacturing methods used to produce the welded material. TODO: ENUM",
        "The mechanical forming and heat treatment methods used to produce the stock material.",
        "Standard designation for the class of coating.",
    ],
)


# process -------------------------------------------------------------

# process/arc_welding_process --------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/process/arc_welding_process",
    asdf_version="1.0.0",
    class_name="ArcWeldingProcess",
    schema_description="See AWS 3.0, Figure 54A - Master Chart of Welding and Allied Processes.",
    properties=["name"],
    required=["name"],
    property_order=["name"],
    property_types=["str"],
    description=["Process name and abbreviation . TODO ENUM"],
)

# process/shielding_gas_for_procedure -------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/process/shielding_gas_for_procedure",
    asdf_version="1.0.0",
    class_name="ShieldingGasForProcedure",
    schema_description="Description of applicable gas composition and flowrates, including torch gas shielding, backing gas, and trailing gas.",
    properties=[
        "use_torch_shielding_gas",
        "torch_shielding_gas",
        "torch_shielding_gas_flowrate",
        "use_backing_gas",
        "backing_gas",
        "backing_gas_flowrate",
        "use_trailing_gas",
        "trailing_shielding_gas",
        "trailing_shielding_gas_flowrate",
    ],
    required=[
        "use_torch_shielding_gas",
        "torch_shielding_gas",
        "torch_shielding_gas_flowrate",
    ],
    property_order=[
        "use_torch_shielding_gas",
        "torch_shielding_gas",
        "torch_shielding_gas_flowrate",
        "use_backing_gas",
        "backing_gas",
        "backing_gas_flowrate",
        "use_trailing_gas",
        "trailing_shielding_gas",
        "trailing_shielding_gas_flowrate",
    ],
    property_types=[
        "bool",
        "ShieldingGasType",
        "pint.Quantity",
        "bool",
        "ShieldingGasType",
        "pint.Quantity",
        "bool",
        "ShieldingGasType",
        "pint.Quantity",
    ],
    description=[
        "Torch shielding gas is/is not required or specified.",
        "Composition of shielding gas expelled from a nozzle in the welding torch.",
        "Flow rate of shielding gas required or specified.",
        "Backing gas is/is not required or specified.",
        "Specification of the component gases of the mixture.",
        "Flowrate of backing gas.",
        "Trailing shielding gas is/is not required or specified.",
        "Composition or identification of trailing gas.",
        "Flowrate of trailing gas during welding.",
    ],
)

# process/gas_component -------------------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/process/gas_component",
    asdf_version="1.0.0",
    class_name="GasComponent",
    schema_description="A single gas element of a mixture and its percentage of the mixture by weight.",
    properties=["gas_chemical_name", "gas_percentage"],
    required=["gas_chemical_name", "gas_percentage"],
    property_order=["gas_chemical_name", "gas_percentage"],
    property_types=["str", "pint.Quantity"],
    description=[
        "Name of a single element or compound of gas. TODO ENUM",
        "Percentage by weight this gas occupies of the total gas mixture.",
    ],
)

# process/shielding_gas_type -------------------------------------------------
create_asdf_dataclass(
    asdf_name="aws/process/shielding_gas_type",
    asdf_version="1.0.0",
    class_name="ShieldingGasType",
    schema_description="Description of a gas or gas mixture used for shielding in arc welding.",
    properties=["gas_component", "common_name", "designation"],
    required=["gas_component", "common_name"],
    property_order=["gas_component", "common_name", "designation"],
    property_types=["List[GasComponent]", "str", "str"],
    description=[
        "A single gas element.",
        "Trade name for the gas mixture.",
        "Specification according to AWS classification by chemical composition of the gas mixture.",
    ],
)



"""Run the AWS Data Dictionary debug example."""

import pprint

import asdf

from weldx.asdf.extension import WeldxAsdfExtension, WeldxExtension
from weldx.asdf.tags.weldx.aws.design.base_metal import BaseMetal
from weldx.asdf.tags.weldx.aws.design.connection import Connection

# weld design -----------------------------------------------------------------
from weldx.asdf.tags.weldx.aws.design.joint_penetration import JointPenetration
from weldx.asdf.tags.weldx.aws.design.sub_assembly import SubAssembly
from weldx.asdf.tags.weldx.aws.design.weld_details import WeldDetails
from weldx.asdf.tags.weldx.aws.design.weldment import Weldment
from weldx.asdf.tags.weldx.aws.design.workpiece import Workpiece
from weldx.asdf.tags.weldx.aws.process.arc_welding_process import ArcWeldingProcess

# welding process -----------------------------------------------------------------
from weldx.asdf.tags.weldx.aws.process.gas_component import GasComponent
from weldx.asdf.tags.weldx.aws.process.shielding_gas_for_procedure import (
    ShieldingGasForProcedure,
)
from weldx.asdf.tags.weldx.aws.process.shielding_gas_type import ShieldingGasType
from weldx.asdf.tags.weldx.core.iso_groove import get_groove
from weldx.constants import WELDX_QUANTITY as Q_

# welding process -----------------------------------------------------------------
gas_comp = [
    GasComponent("argon", Q_(82, "percent")),
    GasComponent("carbon dioxide", Q_(18, "percent")),
]
gas_type = ShieldingGasType(gas_component=gas_comp, common_name="SG")

gas_for_procedure = ShieldingGasForProcedure(
    use_torch_shielding_gas=True,
    torch_shielding_gas=gas_type,
    torch_shielding_gas_flowrate=Q_(20, "l / min"),
)

arc_welding_process = ArcWeldingProcess("GMAW")
process = {
    "arc_welding_process": arc_welding_process,
    "shielding_gas": gas_for_procedure,
}

# weld design -----------------------------------------------------------------
v_groove = get_groove(
    groove_type="VGroove",
    workpiece_thickness=Q_(9, "mm"),
    groove_angle=Q_(50, "deg"),
    root_face=Q_(4, "mm"),
    root_gap=Q_(2, "mm"),
)

joint_penetration = JointPenetration(
    complete_or_partial="completePenetration", root_penetration=Q_(1.0, "mm")
)
weld_details = WeldDetails(
    joint_design=v_groove, weld_sizes=Q_(320, "mm"), number_of_passes=1
)
connection = Connection(
    joint_type="butt_joint",
    weld_type="singleVGroove",
    joint_penetration=joint_penetration,
    weld_details=weld_details,
)
workpieces = [Workpiece(geometry="V-Groove")]
sub_assembly = [SubAssembly(workpiece=workpieces, connection=connection)]

weldment = Weldment(sub_assembly)

base_metal = BaseMetal("steel", "plate", Q_(10.3, "mm"))

filename = "aws_demo.yaml"
tree = dict(process=process, weldment=weldment, base_metal=base_metal)

# Write the data to a new file
with asdf.AsdfFile(
    tree,
    extensions=[WeldxExtension(), WeldxAsdfExtension()],
    ignore_version_mismatch=False,
) as ff:
    ff.write_to(filename, all_array_storage="inline")

# read back data from ASDF file
with asdf.open(
    filename, copy_arrays=True, extensions=[WeldxExtension(), WeldxAsdfExtension()]
) as af:
    data = af.tree
    pprint.pprint(data["process"])
    pprint.pprint(data["weldment"].__dict__)
    pprint.pprint(data["base_metal"].__dict__)



import pandas as pd
import pint

from weldx.asdf.tags.weldx.core.common_types import Variable
from weldx.asdf.types import WeldxType
from weldx.asdf.validators import wx_shape_validator
from weldx.transformations import LocalCoordinateSystem


class LocalCoordinateSystemASDF(WeldxType):
    """Serialization class for weldx.transformations.LocalCoordinateSystem"""

    name = "core/transformations/local_coordinate_system"
    version = "1.0.0"
    types = [LocalCoordinateSystem]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {"wx_shape": wx_shape_validator}

    @classmethod
    def to_tree(cls, node: LocalCoordinateSystem, ctx):
        """
        Convert a 'LocalCoordinateSystem' instance into YAML representations.

        Parameters
        ----------
        node :
            Instance of the 'LocalCoordinateSystem' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the 'LocalCoordinateSystem'
            type to be serialized.

        """
        orientations = Variable(
            "orientations", node.orientation.dims, node.orientation.data
        )

        coordinates = Variable(
            "coordinates", node.coordinates.dims, node.coordinates.data
        )

        tree = {
            "orientations": orientations,
            "coordinates": coordinates,
        }

        if "time" in node.dataset.coords:
            tree["time"] = pd.TimedeltaIndex(node.time)

        if node.reference_time is not None:
            tree["reference_time"] = node.reference_time

        # example code to manipulate inline array storage
        if "time" not in node.coordinates.coords:
            if isinstance(coordinates.data, pint.Quantity):
                ctx.set_array_storage(coordinates.data.magnitude, "inline")
            else:
                ctx.set_array_storage(coordinates.data, "inline")
        if "time" not in node.orientation.coords:
            ctx.set_array_storage(orientations.data, "inline")

        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts basic types representing YAML trees into a 'LocalCoordinateSystem'.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        LocalCoordinateSystem :
            An instance of the 'LocalCoordinateSystem' type.

        """
        orientations = tree["orientations"].data
        coordinates = tree["coordinates"].data

        if "time" in tree:
            time = tree["time"]
        else:
            time = None

        if "reference_time" in tree:
            time_ref = tree["reference_time"]
        else:
            time_ref = None

        return LocalCoordinateSystem(
            orientation=orientations,
            coordinates=coordinates,
            time=time,
            time_ref=time_ref,
        )



from dataclasses import dataclass

import pint

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

__all__ = ["WeldDetails", "WeldDetailsType"]


@dataclass
class WeldDetails:
    """<CLASS DOCSTRING>"""

    joint_design: str
    weld_sizes: pint.Quantity
    number_of_passes: int


class WeldDetailsType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/design/weld_details"
    version = "1.0.0"
    types = [WeldDetails]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: WeldDetails, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = WeldDetails(**tree)
        return obj



"""Test time schema implementation."""
import numpy as np
import pandas as pd
import pytest
from asdf import ValidationError

from weldx.asdf.utils import _write_buffer, _write_read_buffer


@pytest.mark.parametrize(
    "inputs",
    [
        pd.Timedelta("5m3ns"),
        pd.Timedelta("106751 days 23:47:16.854775"),
        pd.timedelta_range(start="-5s", end="25s", freq="3s"),
        pd.TimedeltaIndex([0, 1e9, 5e9, 3e9]),
        pd.Timestamp("2020-04-15T16:47:00.000000001"),
        pd.Timestamp("2020-04-15T16:47:00.000000001", tz="Europe/Berlin"),
        pd.date_range(start="2020-01-01", periods=5, freq="1D"),
        pd.date_range(start="2020-01-01", periods=5, freq="1D", tz="Europe/Berlin"),
        pd.DatetimeIndex(["2020-01-01", "2020-01-02", "2020-01-03", "2020-01-04"]),
        pd.DatetimeIndex(["2020-01-01", "2020-01-02", "2020-01-04", "2020-01-05"]),
    ],
)
def test_time_classes(inputs):
    data = _write_read_buffer({"root": inputs})
    assert np.all(data["root"] == inputs)


def test_time_classes_max_inline():
    with pytest.raises(ValidationError):
        # cannot store large ints >52 bits inline in asdf
        dti = pd.DatetimeIndex(["2020-01-01", "2020-01-02", "2020-01-04", "2020-01-05"])
        _write_buffer(
            {"root": dti}, write_kwargs={"all_array_storage": "inline"},
        )



from dataclasses import dataclass

import pint

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

__all__ = ["GasComponent", "GasComponentType"]


@dataclass
class GasComponent:
    """<CLASS DOCSTRING>"""

    gas_chemical_name: str
    gas_percentage: pint.Quantity


class GasComponentType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/process/gas_component"
    version = "1.0.0"
    types = [GasComponent]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = GasComponent(**tree)
        return obj



from dataclasses import dataclass

import pint

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

__all__ = ["JointPenetration", "JointPenetrationType"]


@dataclass
class JointPenetration:
    """<CLASS DOCSTRING>"""

    complete_or_partial: str
    root_penetration: pint.Quantity
    groove_weld_size: float = None
    incomplete_joint_penetration: float = None
    weld_size: float = None
    weld_size_E1: float = None
    weld_size_E2: float = None
    depth_of_fusion: float = None


class JointPenetrationType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/design/joint_penetration"
    version = "1.0.0"
    types = [JointPenetration]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: JointPenetration, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = JointPenetration(**tree)
        return obj



from dataclasses import dataclass

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

from .joint_penetration import JointPenetration
from .weld_details import WeldDetails

__all__ = ["Connection", "ConnectionType"]


@dataclass
class Connection:
    """<CLASS DOCSTRING>"""

    joint_type: str
    weld_type: str
    joint_penetration: JointPenetration
    weld_details: WeldDetails


class ConnectionType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/design/connection"
    version = "1.0.0"
    types = [Connection]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: Connection, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = Connection(**tree)
        return obj



import re
from typing import Any, Callable, Dict, Iterator, List, Mapping, OrderedDict, Union

import asdf
from asdf import ValidationError
from asdf.schema import _type_to_tag
from asdf.tagged import TaggedDict

from weldx.asdf.extension import WxSyntaxError
from weldx.asdf.tags.weldx.time.datetimeindex import DatetimeIndexType
from weldx.asdf.tags.weldx.time.timedeltaindex import TimedeltaIndexType
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.constants import WELDX_UNIT_REGISTRY as UREG


def _walk_validator(
    instance: OrderedDict,
    validator_dict: OrderedDict,
    validator_function: Callable[[Mapping, Any, str], Iterator[ValidationError]],
    position=None,
    allow_missing_keys: bool = False,
) -> Iterator[ValidationError]:
    """Walk instance and validation dict entries in parallel and apply a validator func.

    This function can be used to recursively walk both the instance dictionary and the
    custom validation dictionary in parallel. Once a leaf dictionary entry is reached,
    the validation function is applied to the selected items.

    Parameters
    ----------
    instance:
        Tree serialization (with default dtypes) of the instance
    validator_dict:
        OrderedDict representation of the validation structure.
    validator_function:
        Custom python validator function to apply along the (nested) dictionary
    position:
        String representation of the current nested schema position
    allow_missing_keys:
        If True will skip validation if the requested key to validate does not exist.

    Yields
    ------
    asdf.ValidationError

    """
    if position is None:  # pragma: no cover
        position = []
    if isinstance(validator_dict, dict):
        for key, item in validator_dict.items():
            if isinstance(item, Mapping):
                yield from _walk_validator(
                    instance[key],
                    validator_dict[key],
                    validator_function,
                    position=position + [key],
                    allow_missing_keys=allow_missing_keys,
                )
            else:
                if key in instance:
                    yield from validator_function(instance[key], item, position + [key])
                elif allow_missing_keys:  # pragma: no cover
                    pass
                else:  # pragma: no cover
                    pass
                    # TODO: if a property is not required the key might be missing
                    # yield ValidationError(f"Missing key {key}")

    else:
        yield from validator_function(instance, validator_dict, position)


def _unit_validator(
    instance: Mapping, expected_dimensionality: str, position: List[str]
) -> Iterator[ValidationError]:
    """Validate the 'unit' key of the instance against the given string.

    Parameters
    ----------
    instance:
        Tree serialization with 'unit' key to validate.
    expected_dimensionality:
        String representation of the unit dimensionality to test against.
    position:
        Current position in nested structure for debugging

    Yields
    ------
    asdf.ValidationError

    """
    if not position:
        position = instance

    unit = instance["unit"]
    valid = Q_(unit).check(UREG.get_dimensionality(expected_dimensionality))
    if not valid:
        yield ValidationError(
            f"Error validating unit dimension for property '{position}'. "
            f"Expected unit of dimension '{expected_dimensionality}' "
            f"but got unit '{unit}'"
        )


def _compare(_int, exp_string):
    """Compare helper of two strings for _custom_shape_validator.

    An integer and an expected string are compared so that the string either contains
    a ":" and thus describes an interval or a string consisting of numbers. So if our
    integer is within the interval or equal to the described number, True is returned.
    The interval can be open, in that there is no number left or right of the ":"
    symbol.

    Examples:
    ---------

    _int = 5
    exp_string = "5"
    -> True

    _int = 5
    exp_string = ":"
    -> True

    Open interval:
    _int = 5
    exp_string = "3:"
    -> True

    _int = 5
    exp_string = "4"
    -> False

    _int = 5
    exp_string = "6:8"
    -> False

    Parameters
    ----------
    _int:
        Integer
    exp_string:
        String with the expected dimension

    Returns
    -------
    bool
        True or False
    """
    if _int < 0:
        raise WxSyntaxError("Negative dimension found")

    if ":" in exp_string:
        ranges = exp_string.split(":")

        if ranges[0] == "":
            ranges[0] = 0
        elif ranges[0].isnumeric():
            ranges[0] = int(ranges[0])
        else:
            raise WxSyntaxError(f"Non numeric character in range {exp_string}")
        if ranges[1] == "":
            ranges[1] = _int
        elif ranges[1].isnumeric():
            ranges[1] = int(ranges[1])
        else:
            raise WxSyntaxError(f"Non numeric character in range {exp_string}")

        if ranges[0] > ranges[1]:
            raise WxSyntaxError(f"The range should not be descending in {exp_string}")
        return int(ranges[0]) <= _int <= int(ranges[1])

    else:
        return _int == int(exp_string)


def _prepare_list(_list, list_expected):
    """Prepare a List and an expected List for validation.

    The preparation of the lists consists in accepting all lists that contain
    white spaces and rewriting all lists that contain the symbols ":" as well as "~" to
    a ":". In addition, lists that begin with "..." or parentheses are reversed for
    validation to work.

    parameters
    ----------
    _list:
        List with values
    list_expected:
        List with expected values
    returns
    -------
    _list:
        prepared List
    list_expected:
        prepared List with expected values
    """
    # remove blank spaces in dict_test
    _list = [x.replace(" ", "") if isinstance(x, str) else x for x in _list]
    # accept "~" additionally as input of ":". And remove blank spaces.
    list_expected = [
        x.replace(" ", "").replace("~", ":") if isinstance(x, str) else x
        for x in list_expected
    ]
    # turn around the list if "..." or "(" are at the beginning.
    # because the validation is made from begin -> end.
    # like this we validate the array from end -> begin.
    if "(" in str(list_expected[0]) or "..." in str(list_expected[0]):
        list_expected = list(reversed(list_expected))
        _list = list(reversed(_list))

    return _list, list_expected


def _is_range_format_valid(format_string: str):
    """
    Return 'True' if a string represents a valid range definition and 'False' otherwise.

    Parameters
    ----------
    format_string:
        String that should be checked.

    Returns
    -------
    bool:
        'True' if the passed string is a valid range definition, 'False' otherwise
    """
    if ":" in format_string:
        if len(format_string.split(":")) != 2:
            return False
        format_string = format_string.replace(":", "")
        return format_string.isalnum() or format_string == ""
    return format_string.isalnum()


def _validate_expected_list(list_expected):
    """Validate an expected List and raises exceptions.

    The syntax of the expected list is validated.
    Examples that will raise errors:

    Variable length should be at beginning or end.
    [1, 2, "...", 4, 5]
    [1, 2, "(3)", 4, 5]

    Additional arguments are not accepted
    [1, 2, 3, 4, "5..."]
    [1, 2, 3, 4, "5(5)"]

    params
    ------
    list_expected:
        Expected List to validate against

    raises
    ------
    ValueError:
        ValueError will be raised if an rule violation is found
    """
    validator = 0
    for exp in list_expected:
        if validator == 1 and not ("(" in str(exp) or "..." in str(exp)):
            raise WxSyntaxError(
                "Optional dimensions in the expected "
                "shape should only stand at the end/beginning."
            )
        if validator == 2:
            raise WxSyntaxError('After "..." should not be another dimension.')
        if "..." in str(exp):
            if "..." != exp:
                raise WxSyntaxError(
                    f'"..." should not have additional properties:' f" {exp} was found."
                )
            validator = 2
        elif "(" in str(exp):
            val = re.search(r"\((.*)\)", exp)
            if (
                val is None
                or len(val.group(1)) + 2 != len(exp)
                or not _is_range_format_valid(val.group(1))
            ):
                raise WxSyntaxError(
                    f'Invalid optional dimension format. Correct format is "(_)", but '
                    f" {exp} was found."
                )

            validator = 1
        elif not _is_range_format_valid(str(exp)):
            raise WxSyntaxError(
                f"{exp} is an invalid range format."
                f"Consult the documentation for a list of all valid options"
            )


def _compare_lists(_list, list_expected):
    """Compare two lists.

    The two lists are interpreted as a list of dimensions. We compare the dimensions of
    these lists and in the case of a mismatch, False is returned. If the dimensions
    match and there are variables in the list_expected, they are entered into a
    dictionary and this is output. The dictionary can be empty if there are no
    variables in the list_expected.

    Examples:
    ---------
    _compare_lists([1, 2, 3], [1, 2, 3])
    -> {}

    _compare_lists([1, 2, 3], [1, n1, n2])
    -> {n1: 2, n2: 3}

    _compare_lists([1, 2, 3], [1, "..."])
    -> {}

    _compare_lists([1, 2, 3], [1, 2, 4])
    -> False

    params
    ------
    _list:
        List of Integer
    list_expected:
        List build by the rules in _custom_shape_validator
    returns
    -------
    False:
        when a dimension mismatch occurs
    dict_values:
        when no dimension mismatch occurs. Can be empty {}. Dictionary - keys: variable
        names in the validation schemes. values: values of the validation schemes.
    """
    dict_values = dict()

    has_variable_dim_num = False
    for i, exp in enumerate(list_expected):

        if "..." in str(exp):
            has_variable_dim_num = True
            break  # all the following dimensions are accepted

        if "(" in str(exp):
            if i < len(_list):
                exp = re.search(r"\((.*)\)", exp).group(1)
            else:  # pragma: no cover
                continue  # TODO: actually covered, but not registered by codecov - bug?

        # all alphanumeric strings are OK - only numeric strings are not
        # eg: "n", "n1", "n1234", "myasdfstring1337"
        if str(exp).isalnum() and not str(exp).isnumeric():
            if exp not in dict_values:
                dict_values[exp] = _list[i]
            elif _list[i] != dict_values[exp]:
                return False

        elif i >= len(_list) or not _compare(_list[i], str(exp)):
            return False

    if (len(_list) > len(list_expected)) and not has_variable_dim_num:
        return False

    return dict_values


def _get_instance_shape(instance_dict: Union[TaggedDict, Dict[str, Any]]) -> List[int]:
    """Get the shape of an ASDF instance from its tagged dict form."""
    if isinstance(instance_dict, (float, int)):  # test against [1] for single values
        return [1]
    elif "shape" in instance_dict:
        return instance_dict["shape"]
    elif isinstance(instance_dict, asdf.types.tagged.Tagged):
        # add custom type implementations
        if "weldx/time/timedeltaindex" in instance_dict._tag:
            return TimedeltaIndexType.shape_from_tagged(instance_dict)
        elif "weldx/time/datetimeindex" in instance_dict._tag:
            return DatetimeIndexType.shape_from_tagged(instance_dict)

    return None


def _custom_shape_validator(dict_test: Dict[str, Any], dict_expected: Dict[str, Any]):
    """Validate dimensions which are stored in two dictionaries dict_test and
    dict_expected.

    Syntax for the dict_expected:
    -----------------------------
    Items with arrays with each value having the following Syntax:
    1)  3 : an integer indicates a fix dimension for the same item in dict_test
    2)  "~", ":" or None : this string indicates a single dimension of arbitrary length.
    3)  "..." : this string indicates an arbitrary number of dimensions of arbitrary
            length. Can be optional.
    4)  "2~4" : this string indicates a single dimension of length 2, 3 or 4. This
            has to be ascending or you can give an unlimited interval limit like this
            "2~" which would indicate a single dimension of length greater then 1.
    5)  "n" : this indicates a single dimension fixed to a letter. Any letter or
            combination of letters should work The letter will be compared to the same
            letter in all the arrays in the dict_expected.
    6)  (x) : parenthesis indicates that the dict_test does not need to have this
            dimension. This can NOT be combined with 3) or the None from 2).

    Parameters
    ----------
    dict_test:
        dictionary to test against
    dict_expected:
        dictionary with the expected values

    raises
    ------
    ValueError:
        when dict_expected does violate against the Syntax rules

    returns
    -------
    False
        when any dimension mismatch occurs
    dict_values
        Dictionary - keys: variable names in the validation schemes. values: values of
        the validation schemes.
    """

    dict_values = {}

    # catch single shape definitions
    if isinstance(dict_expected, list):
        # get the shape of current
        shape = _get_instance_shape(dict_test)
        if not shape:
            raise ValidationError(f"Could not determine shape in instance {dict_test}.")

        list_test, list_expected = _prepare_list(shape, dict_expected)

        _validate_expected_list(list_expected)
        _dict_values = _compare_lists(list_test, list_expected)

        if _dict_values is False:
            raise ValidationError(
                f"Shape {list_test[::-1]} does not match requirement "
                f"{list_expected[::-1]}"
            )

        return _dict_values

    elif isinstance(dict_expected, dict):
        for key, item in dict_expected.items():
            # allow optional syntax
            _optional = False
            if key.startswith("(") and key.endswith(")"):
                key = key[1:-1]
                _optional = True
                if len(key) == 0:
                    raise WxSyntaxError("wx_shape entry undefined")

            # test shapes
            if key in dict_test:
                # go one level deeper in the dictionary
                _dict_values = _custom_shape_validator(dict_test[key], item)
            elif _optional:
                _dict_values = {}
            else:
                raise ValidationError(
                    f"Could not access key '{key}'  in instance {dict_test}."
                )

            for key in _dict_values:
                if key not in dict_values:
                    dict_values[key] = _dict_values[key]
                elif dict_values[key] != _dict_values[key]:
                    return False
    else:
        raise WxSyntaxError(
            f"Found an incorrect object: {type(dict_expected)}. "
            "Should be a dict or list."
        )

    return dict_values


def wx_unit_validator(
    validator, wx_unit, instance, schema
) -> Iterator[ValidationError]:
    """Custom validator for checking dimensions for objects with 'unit' property.

    ASDF documentation:
    https://asdf.readthedocs.io/en/2.6.0/asdf/extensions.html#adding-custom-validators

    Parameters
    ----------
    validator:
        A jsonschema.Validator instance.
    wx_unit:
        Enable unit validation for this schema.
    instance:
        Tree serialization (with default dtypes) of the instance
    schema:
        Dict representing the full ASDF schema.

    Yields
    ------
    asdf.ValidationError

    """
    yield from _walk_validator(
        instance=instance,
        validator_dict=wx_unit,
        validator_function=_unit_validator,
        position=[],
        allow_missing_keys=False,
    )


def wx_shape_validator(
    validator, wx_shape, instance, schema
) -> Iterator[ValidationError]:
    """Custom validator for checking dimensions for objects with 'shape' property.

    ASDF documentation:
    https://asdf.readthedocs.io/en/2.6.0/asdf/extensions.html#adding-custom-validators

    Parameters
    ----------
    validator:
        A jsonschema.Validator instance.
    wx_shape:
        Enable shape validation for this schema.
    instance:
        Tree serialization (with default dtypes) of the instance
    schema:
        Dict representing the full ASDF schema.

    Yields
    ------
    asdf.ValidationError

    """

    dim_dict = None
    try:
        dim_dict = _custom_shape_validator(instance, wx_shape)
    except ValidationError:
        yield ValidationError(
            f"Error validating shape {wx_shape}.\nOn instance {instance}"
        )

    if isinstance(dim_dict, dict):
        return None
    else:
        yield ValidationError(
            f"Error validating shape {wx_shape}.\nOn instance {instance}"
        )


def _compare_tag_version(instance_tag: str, tagname: str):
    """Compare ASDF tag-strings with flexible version syntax.

    Parameters
    ----------
    instance_tag:
        the full ASDF tag to validate
    tagname:
        tag string with custom version syntax to validate against

    Returns
    -------
        bool
    """
    if instance_tag is None:
        return True

    if instance_tag.startswith("tag:yaml.org"):  # test for python builtins
        return instance_tag == tagname
    instance_tag_version = [int(v) for v in instance_tag.rpartition("-")[-1].split(".")]

    tag_parts = tagname.rpartition("-")
    tag_uri = tag_parts[0]
    tag_version = [v for v in tag_parts[-1].split(".")]

    if tag_version == ["*"]:
        version_compatible = True
    elif all([vstr.isdigit() for vstr in tag_version]):
        vnum = [int(vstr) for vstr in tag_version]
        version_compatible = all(
            [v[0] == v[1] for v in zip(vnum, instance_tag_version)]
        )
    else:
        raise WxSyntaxError(f"Unknown wx_tag syntax {tagname}")

    if (not instance_tag.startswith(tag_uri)) or (not version_compatible):
        return False
    return True


def wx_tag_validator(validator, tagname, instance, schema):
    """Validate instance tag string with flexible version syntax.

    The following syntax is allowed to validate against:

    wx_tag: http://stsci.edu/schemas/asdf/core/software-* # allow every version
    wx_tag: http://stsci.edu/schemas/asdf/core/software-1 # fix major version
    wx_tag: http://stsci.edu/schemas/asdf/core/software-1.2 # fix minor version
    wx_tag: http://stsci.edu/schemas/asdf/core/software-1.2.3 # fix patch version

    Parameters
    ----------
    validator:
        A jsonschema.Validator instance.
    tagname:
        tag string with custom version syntax to validate against
    instance:
        Tree serialization (with default dtypes) of the instance
    schema:
        Dict representing the full ASDF schema.

    Returns
    -------
        bool

    """
    if hasattr(instance, "_tag"):
        instance_tag = instance._tag
    else:
        # Try tags for known Python builtins
        instance_tag = _type_to_tag(type(instance))

    if instance_tag is not None:
        if not _compare_tag_version(instance_tag, tagname):
            yield ValidationError(
                "mismatched tags, wanted '{0}', got '{1}'".format(tagname, instance_tag)
            )


def wx_property_tag_validator(
    validator, wx_property_tag: str, instance, schema
) -> Iterator[ValidationError]:
    """

    Parameters
    ----------
    validator
        A jsonschema.Validator instance.
    wx_property_tag
        The tag to test all object properties against.
    instance
        Tree serialization (with default dtypes or as tagged dict) of the instance
    schema
        Dict representing the full ASDF schema.

    Yields
    ------
    asdf.ValidationError

    """
    for _, value in instance.items():
        yield from wx_tag_validator(
            validator, tagname=wx_property_tag, instance=value, schema=None
        )



import numpy as np
from scipy.spatial.transform import Rotation

from weldx.asdf.types import WeldxType
from weldx.asdf.validators import wx_unit_validator
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.transformations import WXRotation


class WXRotationTypeASDF(WeldxType):
    """Serialization class for the 'Scipy.Rotation' type"""

    name = "core/transformations/rotation"
    version = "1.0.0"
    types = [Rotation]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {"wx_unit": wx_unit_validator}

    @classmethod
    def to_tree(cls, node: Rotation, ctx):
        """
        Convert an instance of the 'Dimension' type into YAML representations.

        Parameters
        ----------
        node :
            Instance of the 'Dimension' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the 'Dimension' type to be
            serialized.

        """
        tree = {}

        if not hasattr(node, "wx_meta"):  # default to quaternion representation
            tree["quaternions"] = node.as_quat()
        elif node.wx_meta["constructor"] == "from_quat":
            tree["quaternions"] = node.as_quat()
        elif node.wx_meta["constructor"] == "from_matrix":
            tree["matrix"] = node.as_matrix()
        elif node.wx_meta["constructor"] == "from_rotvec":
            tree["rotvec"] = node.as_rotvec()
        elif node.wx_meta["constructor"] == "from_euler":
            seq_str = node.wx_meta["seq"]
            if not len(seq_str) == 3:
                if all([c in "xyz" for c in seq_str]):
                    seq_str = seq_str + "".join([c for c in "xyz" if c not in seq_str])
                elif all([c in "XYZ" for c in seq_str]):
                    seq_str = seq_str + "".join([c for c in "XYZ" if c not in seq_str])
                else:  # pragma: no cover
                    raise ValueError("Mix of intrinsic and extrinsic euler angles.")

            angles = node.as_euler(seq_str, degrees=node.wx_meta["degrees"])
            angles = np.squeeze(angles[..., : len(node.wx_meta["seq"])])

            if node.wx_meta["degrees"]:
                angles = Q_(angles, "degree")
            else:
                angles = Q_(angles, "rad")

            tree["sequence"] = node.wx_meta["seq"]
            tree["angles"] = angles

        else:  # pragma: no cover
            raise NotImplementedError("unknown or missing constructor")

        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts basic types representing YAML trees into custom types.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        Dimension :
            An instance of the 'Dimension' type.

        """
        if "quaternions" in tree:
            return WXRotation.from_quat(tree["quaternions"])
        elif "matrix" in tree:
            return WXRotation.from_matrix(tree["matrix"])
        elif "rotvec" in tree:
            return WXRotation.from_rotvec(tree["rotvec"])
        elif "angles" in tree:
            if "degree" in str(tree["angles"].units):
                angles = tree["angles"].to("degree").magnitude
                degrees = True
            else:
                angles = tree["angles"].to("rad").magnitude
                degrees = False
            return WXRotation.from_euler(
                seq=tree["sequence"], angles=angles, degrees=degrees
            )



from dataclasses import dataclass

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

__all__ = ["ArcWeldingProcess", "ArcWeldingProcessType"]

_name_to_abbr = {
    "atomicHydrogenWelding": "AHW",
    "bareMetalArcWelding": "BMAW",
    "carbonArcWelding": "CAW",
    "carbonArcWeldingGas": "CAW-G",
    "carbonArcWeldingShielded": "CAW-S",
    "electrogasWelding": "EGW",
    "electroSlagWelding": "ESW",
    "gasMetalArcWelding": "GMAW",
    "gasTungstenArcWelding": "GTAW",
    "plasmaArcWelding": "PAW",
    "shieldedMetalArcWelding": "SMAW",
    "studArcWelding": "SW",
    "submergedArcWelding": "SAW",
    "submergedArcWeldingSeries": "SAW-S",
}

_abbr_to_name = {
    "AHW": "atomicHydrogenWelding",
    "BMAW": "bareMetalArcWelding",
    "CAW": "carbonArcWelding",
    "CAW-G": "carbonArcWeldingGas",
    "CAW-S": "carbonArcWeldingShielded",
    "EGW": "electrogasWelding",
    "ESW": "electroSlagWelding",
    "GMAW": "gasMetalArcWelding",
    "GTAW": "gasTungstenArcWelding",
    "PAW": "plasmaArcWelding",
    "SMAW": "shieldedMetalArcWelding",
    "SW": "studArcWelding",
    "SAW": "submergedArcWelding",
    "SAW-S": "submergedArcWeldingSeries",
}


@dataclass
class ArcWeldingProcess:
    """<CLASS DOCSTRING>"""

    name: str
    abbreviation: str

    def __init__(self, name_or_abbreviation):
        if name_or_abbreviation in _name_to_abbr:
            self.name = name_or_abbreviation
            self.abbreviation = _name_to_abbr[name_or_abbreviation]
        elif name_or_abbreviation in _abbr_to_name:
            self.name = _abbr_to_name[name_or_abbreviation]
            self.abbreviation = name_or_abbreviation
        else:
            raise ValueError(
                f"Could not find process matching description '{name_or_abbreviation}'"
            )


class ArcWeldingProcessType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/process/arc_welding_process"
    version = "1.0.0"
    types = [ArcWeldingProcess]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = ArcWeldingProcess(tree["name"])
        return obj



"""Temporary module for welding related classes."""

from dataclasses import dataclass
from typing import Dict

from asdf.tagged import tag_object

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.core import TimeSeries


@dataclass
class GmawProcess:
    """Container class for all GMAW processes."""

    base_process: str
    manufacturer: str
    power_source: str
    parameters: Dict[str, TimeSeries]
    tag: str = None
    meta: dict = None

    def __post_init__(self):
        """Set defaults and convert parmater inputs."""
        if self.tag is None:
            self.tag = "GMAW"

        self.parameters = {
            k: (v if isinstance(v, TimeSeries) else TimeSeries(v))
            for k, v in self.parameters.items()
        }


class GmawProcessTypeAsdf(WeldxType):
    """Custom serialization class for GmawProcess."""

    name = ["process/GMAW", "process/CLOOS/spray_arc", "process/CLOOS/pulse"]
    version = "1.0.0"
    types = [GmawProcess]
    requires = ["weldx"]

    @classmethod
    def to_tree(cls, node: GmawProcess, ctx):
        """Convert tree and remove all None entries from node dictionary."""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def to_tree_tagged(cls, node: GmawProcess, ctx):
        """Serialize tree with custom tag definition."""
        tree = cls.to_tree(node, ctx)
        tag = "tag:weldx.bam.de:weldx/process/" + tree["tag"] + "-" + str(cls.version)
        return tag_object(tag, tree, ctx=ctx)

    @classmethod
    def from_tree(cls, tree, ctx):
        """Read tree object into dataclass."""
        obj = GmawProcess(**tree)
        return obj



import warnings

# asdf extensions and tags
import weldx.asdf

# main packages
import weldx.core
import weldx.geometry
import weldx.transformations
import weldx.utility
import weldx.welding

# class imports to weldx namespace
from weldx.asdf.tags.weldx.core.iso_groove import get_groove
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.core import MathematicalExpression, TimeSeries
from weldx.geometry import ArcSegment, Geometry, LineSegment, Profile, Shape, Trace
from weldx.transformations import (
    CoordinateSystemManager,
    LocalCoordinateSystem,
    WXRotation,
)

# versioneer
from ._version import get_versions

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    Q_([])

__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "core",
    "geometry",
    "measurement",
    "transformations",
    "utility",
    "asdf",
    "Q_",
    "welding",
]



"""Test custom weldx ASDF validator functions."""
import numpy as np
import pandas as pd
import pytest
from asdf import ValidationError

from weldx import Q_
from weldx.asdf.extension import WxSyntaxError
from weldx.asdf.tags.weldx.debug.test_property_tag import PropertyTagTestClass
from weldx.asdf.tags.weldx.debug.test_shape_validator import ShapeValidatorTestClass
from weldx.asdf.tags.weldx.debug.test_unit_validator import UnitValidatorTestClass
from weldx.asdf.utils import _write_read_buffer
from weldx.asdf.validators import _compare_tag_version, _custom_shape_validator


@pytest.mark.parametrize(
    "instance_tag,tagname,result",
    [
        (None, "tag:debug.com/object-*", True),
        ("tag:debug.com/object-1.2.3", "tag:debug.com/object-*", True),
        ("http://debug.com/object-1.2.3", "http://debug.com/object-*", True),
        ("http://debug.com/object-1.2.3", "http://debug.com/object-1.2.3", True),
        ("http://debug.com/object-1.2.3", "http://debug.com/object-1.2", True),
        ("http://debug.com/object-1.2.3", "http://debug.com/object-1", True),
        ("http://debug.com/object-1.2.3", "http://debug.com/object-2", False),
        ("http://debug.com/object-2.0.0", "http://debug.com/object-1", False),
        ("http://debug.com/object-2.0.0", "http://debug.com/object-2.1", False),
        ("http://debug.com/object-2.0.0", "http://debug.com/other-2.0.0", False),
        ("http://debug.com/object-2.0.0", "http://other.com/object-2.0.0", False),
        ("http://debug.com/object-1.2.3", "http://other.com/object-1.2.3", False),
    ],
)
def test_wx_tag_syntax(instance_tag, tagname, result):
    """Test ASDF tag version syntax resolving."""
    assert _compare_tag_version(instance_tag, tagname) == result


@pytest.mark.parametrize(
    "instance_tag,tagname,err",
    [
        ("tag:debug.com/object-1.2.3", "tag:debug.com/object", WxSyntaxError),
        ("tag:debug.com/object-1.2.3", "tag:debug.com/object-", WxSyntaxError),
        ("tag:debug.com/object-1.2.3", "tag:debug.com/object-**", WxSyntaxError),
    ],
)
def test_wx_tag_syntax_exceptions(instance_tag, tagname, err):
    """Test custom ASDF shape validators."""
    with pytest.raises(err):
        _compare_tag_version(instance_tag, tagname)


@pytest.mark.parametrize(
    "test_input", [PropertyTagTestClass()],
)
def test_property_tag_validator(test_input):
    """Test custom ASDF shape validators."""
    _write_read_buffer({"root_node": test_input})


@pytest.mark.parametrize(
    "test_input,err",
    [
        (PropertyTagTestClass(prop3=pd.Timedelta(2, "s")), ValidationError),
        (PropertyTagTestClass(prop3="STRING"), ValidationError),
    ],
)
def test_property_tag_validator_exceptions(test_input, err):
    """Test custom ASDF shape validators."""
    with pytest.raises(err):
        _write_read_buffer({"root_node": test_input})


def _val(list_test, list_expected):
    """Add shape key to lists."""
    if isinstance(list_test, list):
        res = _custom_shape_validator({"shape": list_test}, list_expected)
        return isinstance(res, dict)
    return isinstance(_custom_shape_validator(list_test, list_expected), dict)


@pytest.mark.parametrize(
    "shape, exp",
    [
        ([3], [3]),
        ([2, 4, 5], [2, 4, 5]),
        ([1, 2, 3], ["..."]),
        ([1, 2], [1, 2, "..."]),
        ([1, 2], ["...", 1, 2]),
        ([1, 2, 3], [1, 2, None]),
        ([1, 2, 3], [None, 2, 3]),
        ([1], [1, "..."]),
        ([1, 2, 3, 4, 5], [1, "..."]),
        ([1, 2, 3, 4, 5], ["...", 4, 5]),
        ([1, 2], [1, 2, "(3)"]),
        ([1, 2], [1, 2, "(n)"]),
        ([1, 2], [1, 2, "(2)", "(3)"]),
        ([2, 3], ["(1)", 2, 3]),
        ([1, 2, 3], ["(1)", 2, 3]),
        ([2, 3], ["(1~3)", 2, 3]),
        ([2, 2, 3], ["(1~3)", 2, 3]),
        ([1, 2, 3], [1, "1~3", 3]),
        ([1, 2, 3], [1, "1~", 3]),
        ([1, 2, 3], [1, "~3", 3]),
        ([1, 2, 3], [1, "~", 3]),
        ([1, 200, 3], [1, "~", 3]),
        ([1, 2, 3], [1, 2, "(~)"]),
        ([1, 2, 300], [1, 2, "(~)"]),
        ([1, 2, 3], [1, "(n)", "..."]),
        (1.0, [1]),
    ],
)
def test_shape_validator_syntax2(shape, exp):
    assert _val(shape, exp)


@pytest.mark.parametrize(
    "shape, exp, err",
    [
        ([2, 2, 3], [1, "..."], ValidationError),
        ([2, 2, 3], ["...", 1], ValidationError),
        ([1], [1, 2], ValidationError),
        ([1, 2], [1], ValidationError),
        ([1, 2], [3, 2], ValidationError),
        ([1], [1, "~"], ValidationError),
        ([1], ["~", 1], ValidationError),
        ([1, 2, 3], [1, 2, "(4)"], ValidationError),
        ([1, 2, 3], ["(2)", 2, 3], ValidationError),
        ([1, 2], [1, "4~8"], ValidationError),
        ([1, 9], [1, "4~8"], ValidationError),
        ([1, 2], [1, "(4~8)"], ValidationError),
        ([1, 9], [1, "(4~8)"], ValidationError),
        (1.0, [2], ValidationError),
        ([1, 2, 3, 4], [1, 2, "n", "n"], ValidationError),
        ([1, 2], [1, "~", "(...)"], WxSyntaxError),
        ([1, 2], [1, "(2)", 3], WxSyntaxError),
        ([1, 2], [1, 2, "((3))"], WxSyntaxError),
        ([1, 2], [1, 2, "3)"], WxSyntaxError),
        ([1, 2], [1, 2, "*3"], WxSyntaxError),
        ([1, 2], [1, 2, "(3"], WxSyntaxError),
        ([1, 2], [1, 2, "(3)3"], WxSyntaxError),
        ([1, 2], [1, 2, "2(3)"], WxSyntaxError),
        ([1, 2], [1, "...", 2], WxSyntaxError),
        ([1, 2], ["(1)", "..."], WxSyntaxError),
        ([1, 2], [1, "4~1"], WxSyntaxError),
        ([-1, -2], [-1, -2], WxSyntaxError),
        ([-1, 2], [1, 2], WxSyntaxError),
        ([1, 2], [-1, 2], WxSyntaxError),
        ([1, 2], [1, 2, "(-3)"], WxSyntaxError),
        ([1, 2], [1, 2, "(-3~-1)"], WxSyntaxError),
        ([1, 2], [1, 2, "(-3~1)"], WxSyntaxError),
        ([1, 2, 1], ["(-3~1)", 2, 1], WxSyntaxError),
        ([1, 2], [1, "(9~m)"], WxSyntaxError),
        ([1, 2], [1, "(n~9)"], WxSyntaxError),
        ([1, 2], [1, "(n~m)"], WxSyntaxError),
        ([1, 2], [1, "(1~3~5)"], WxSyntaxError),
        ("a string", [1, "(1~3~5)"], ValidationError),
        ([1, 2], "a string", WxSyntaxError),
    ],
)
def test_shape_validation_error_exception(shape, exp, err):
    with pytest.raises(err):
        assert _val(shape, exp)


@pytest.mark.parametrize(
    "test_input",
    [
        ShapeValidatorTestClass(),
        ShapeValidatorTestClass(time_prop=pd.date_range("2020", freq="D", periods=9)),
        ShapeValidatorTestClass(optional_prop=np.ones((1, 2, 3)),),
        ShapeValidatorTestClass(
            nested_prop={
                "p1": np.ones((10, 8, 6, 4, 2)),
                "p2": np.ones((9, 7, 5, 3, 1)),
                "p3": np.ones((1, 2, 3)),
            }
        ),
    ],
)
def test_shape_validator(test_input):
    _write_read_buffer({"root": test_input})


@pytest.mark.parametrize(
    "test_input",
    [
        ShapeValidatorTestClass(
            prop4=np.ones((2, 3, 5, 7, 9)),  # mismatch a with prop5
        ),
        ShapeValidatorTestClass(prop2=np.ones((5, 2, 1)),),  # mismatch n with prop1
        ShapeValidatorTestClass(
            nested_prop={"p1": np.ones((10, 8, 6, 4, 2))},  # missing p2
        ),
        ShapeValidatorTestClass(optional_prop=np.ones((3, 2, 9)),),  # wrong optional
        ShapeValidatorTestClass(time_prop=pd.date_range("2020", freq="D", periods=3)),
    ],
)
def test_shape_validator_exceptions(test_input):
    with pytest.raises(ValidationError):
        _write_read_buffer({"root": test_input})


@pytest.mark.parametrize(
    "test",
    [
        UnitValidatorTestClass(
            length_prop=Q_(1, "inch"),
            velocity_prop=Q_(2, "km / s"),
            current_prop=Q_(np.eye(2, 2), "mA"),
            nested_prop=dict(q1=Q_(np.eye(3, 3), "m"), q2=Q_(2, "m^3")),
            simple_prop={"value": float(3), "unit": "m"},
        ),
        pytest.param(
            UnitValidatorTestClass(
                length_prop=Q_(1, "s"),  # wrong unit
                velocity_prop=Q_(2, "km / s"),
                current_prop=Q_(np.eye(2, 2), "mA"),
                nested_prop=dict(q1=Q_(np.eye(3, 3), "m"), q2=Q_(2, "m^3")),
                simple_prop={"value": float(3), "unit": "m"},
            ),
            marks=pytest.mark.xfail(raises=ValidationError),
        ),
        pytest.param(
            UnitValidatorTestClass(
                length_prop=Q_(1, "s"),
                velocity_prop=Q_(2, "liter"),  # wrong unit
                current_prop=Q_(np.eye(2, 2), "mA"),
                nested_prop=dict(q1=Q_(np.eye(3, 3), "m"), q2=Q_(2, "m^3")),
                simple_prop={"value": float(3), "unit": "m"},
            ),
            marks=pytest.mark.xfail(raises=ValidationError),
        ),
        pytest.param(
            UnitValidatorTestClass(
                length_prop=Q_(1, "inch"),
                velocity_prop=Q_(2, "km / s"),
                current_prop=Q_(np.eye(2, 2), "V"),  # wrong unit
                nested_prop=dict(q1=Q_(np.eye(3, 3), "m"), q2=Q_(2, "m^3")),
                simple_prop={"value": float(3), "unit": "m"},
            ),
            marks=pytest.mark.xfail(raises=ValidationError),
        ),
        pytest.param(
            UnitValidatorTestClass(
                length_prop=Q_(1, "m"),
                velocity_prop=Q_(2, "km / s"),
                current_prop=Q_(np.eye(2, 2), "mA"),
                nested_prop=dict(q1=Q_(np.eye(3, 3), "m"), q2=Q_(2, "V")),  # wrong unit
                simple_prop={"value": float(3), "unit": "m"},
            ),
            marks=pytest.mark.xfail(raises=ValidationError),
        ),
        pytest.param(
            UnitValidatorTestClass(
                length_prop=Q_(1, "m"),
                velocity_prop=Q_(2, "km / s"),
                current_prop=Q_(np.eye(2, 2), "mA"),
                nested_prop=dict(q1=Q_(np.eye(3, 3), "m"), q2=Q_(2, "m^3")),
                simple_prop={"value": float(3), "unit": "s"},  # wrong unit
            ),
            marks=pytest.mark.xfail(raises=ValidationError),
        ),
    ],
)
def test_unit_validator(test):
    data = _write_read_buffer({"root_node": test})
    test_read = data["root_node"]
    assert isinstance(data, dict)
    assert test_read.length_prop == test.length_prop
    assert test_read.velocity_prop == test.velocity_prop
    assert np.all(test_read.current_prop == test.current_prop)
    assert np.all(test_read.nested_prop["q1"] == test.nested_prop["q1"])
    assert test_read.nested_prop["q2"] == test.nested_prop["q2"]
    assert test_read.simple_prop == test.simple_prop



"""Tests of the core package."""


import numpy as np
import pandas as pd
import pint
import pytest

import weldx.utility as ut
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.constants import WELDX_UNIT_REGISTRY as UREG
from weldx.core import MathematicalExpression, TimeSeries


# Todo: Move this to conftest.py?
def get_test_name(param):
    """Get the test name from the parameter list of a parametrized test."""
    if isinstance(param, str) and param[0] == "#":
        return param[1:]
    return ""


# --------------------------------------------------------------------------------------
# MathematicalExpression
# --------------------------------------------------------------------------------------


class TestMathematicalExpression:
    """Tests the mathematical expression class."""

    # Fixtures, aliases and shared variables -------------------------------------------

    ME = MathematicalExpression
    # unfortunately, fixtures can not be used in a parametrize section
    expr_def = "(a + b)**2 + c - d"
    params_def = {"a": 2, "c": 3.5}

    @staticmethod
    @pytest.fixture()
    def ma_def() -> MathematicalExpression:
        """Get a default instance for tests."""
        return MathematicalExpression(
            TestMathematicalExpression.expr_def, TestMathematicalExpression.params_def,
        )

    # Helper functions -----------------------------------------------------------------

    @staticmethod
    def _check_params_and_vars(expression, exp_params, exp_vars):
        """Check parameters and variables of an MathematicalExpression."""
        assert expression.num_parameters == len(exp_params)
        assert expression.parameters == exp_params

        assert expression.num_variables == len(exp_vars)
        for variable in exp_vars:
            assert variable in expression.get_variable_names()

    # test_construction ----------------------------------------------------------------

    @pytest.mark.parametrize(
        "expression, parameters,  exp_vars",
        [
            ("a*b + c/d - e", {"d": 1, "e": 2}, ["a", "b", "c"]),
            ("a*b + c/d - e", {}, ["a", "b", "c", "d", "e"]),
            ("a**2 + b - c", {"a": 1, "c": 2}, ["b"]),
        ],
    )
    def test_construction(self, expression, parameters, exp_vars):
        """Test the construction."""
        expr = MathematicalExpression(expression=expression, parameters=parameters)

        self._check_params_and_vars(expr, parameters, exp_vars)

    # test_construction_exceptions -----------------------------------------------------

    @staticmethod
    @pytest.mark.parametrize(
        "expression, parameters, exception_type, name",
        [
            ("a*b + c/d - e", {"f": 1}, ValueError, "# parameter not in expression"),
            ("a*b + c/d - e", 1, ValueError, "# invalid parameter type"),
            ("a + $b#!==3", {"a": 1}, Exception, "# invalid expression"),
        ],
        ids=get_test_name,
    )
    def test_construction_exceptions(expression, parameters, exception_type, name):
        """Test the exceptions of the '__init__' method."""
        with pytest.raises(exception_type):
            MathematicalExpression(expression=expression, parameters=parameters)

    # test_set_parameter ---------------------------------------------------------------

    def test_set_parameter(self):
        """Test the set_parameter function of the mathematical expression."""
        expr = MathematicalExpression("a*b + c/d - e")

        # check initial configuration
        self._check_params_and_vars(expr, {}, ["a", "b", "c", "d", "e"])

        # set first parameters
        expr.set_parameter("d", 1)
        expr.set_parameter("e", 2)

        self._check_params_and_vars(expr, {"d": 1, "e": 2}, ["a", "b", "c"])

        # set another parameter and overwrite others
        expr.set_parameter("a", 5)
        expr.set_parameter("d", 7)
        expr.set_parameter("e", -1)

        self._check_params_and_vars(expr, {"a": 5, "d": 7, "e": -1}, ["b", "c"])

    # test_set_parameter_exceptions ----------------------------------------------------

    @staticmethod
    @pytest.mark.parametrize(
        "name, value, exception_type, test_name",
        [
            ("k", 1, ValueError, "# parameter not in expression"),
            (33, 1, TypeError, "# wrong type as name #1"),
            ({"a": 1}, 1, TypeError, "# wrong type as name #2"),
        ],
        ids=get_test_name,
    )
    def test_set_parameter_exceptions(ma_def, name, value, exception_type, test_name):
        """Test the exceptions of the 'set_parameter' method."""
        with pytest.raises(exception_type):
            ma_def.set_parameter(name, value)

    # test_comparison ------------------------------------------------------------------

    expr_mat_identical = "a**2 + 2*a*b + b**2 + c - d"
    expr_different = "a*b + c*d"
    params_too_many = {"a": 2, "c": 3.5, "d": 4}
    params_wrong_value = {"a": 2, "c": 1.5}

    @staticmethod
    @pytest.mark.parametrize(
        "other, equal, equal_no_params, mat_equal, mat_equal_no_params",
        [
            (ME(expr_def, params_def), True, True, True, True),
            (ME(expr_mat_identical, params_def), False, False, True, True),
            (ME(expr_different, params_def), False, False, False, False),
            (ME(expr_def, params_too_many), False, True, False, True),
            (ME(expr_mat_identical, params_too_many), False, False, False, True),
            (ME(expr_different, params_too_many), False, False, False, False),
            (ME(expr_def, params_wrong_value), False, True, False, True),
            (ME(expr_mat_identical, params_wrong_value), False, False, False, True),
            (ME(expr_different, params_wrong_value), False, False, False, False),
            (1, False, False, False, False),
            ("I am not a MathematicalExpression", False, False, False, False),
        ],
    )
    def test_comparison(
        ma_def, other, equal, equal_no_params, mat_equal, mat_equal_no_params,
    ):
        """Test if another object is equal to the default instance."""
        assert (ma_def == other) is equal
        assert (ma_def != other) is not equal
        assert ma_def.equals(other, False, True) is equal_no_params
        assert ma_def.equals(other) is mat_equal
        assert ma_def.equals(other, False, False) is mat_equal_no_params

    # -----------------------------------------------------

    # TODO: Add tests for quantities
    @staticmethod
    @pytest.mark.parametrize(
        "expression, parameters, variables, exp_result",
        [
            ("a*b + c/d - e", {"d": 1, "e": 2}, {"a": 1, "b": 2, "c": 3}, 3),
            ("(a + b)**2 + c - d", {"a": 3, "d": 2}, {"b": 2, "c": 4}, 27),
            ("a + b", {"a": np.array([1, 2])}, {"b": np.array([2, 4])}, [3, 6],),
        ],
    )
    def test_evaluation(expression, parameters, variables, exp_result):
        """Test the evaluation of the mathematical function."""
        expr = MathematicalExpression(expression=expression, parameters=parameters)

        assert np.all(expr.evaluate(**variables) == exp_result)

    # test_evaluate_exceptions ---------------------------------------------------------

    @staticmethod
    @pytest.mark.parametrize(
        "variables, exception_type, test_name",
        [
            ({"b": 1, "c": 2, "d": 3}, ValueError, "# input is expression parameter"),
            ({"b": 1}, Exception, "# not enough values provided"),
        ],
        ids=get_test_name,
    )
    def test_evaluate_exceptions(ma_def, variables, exception_type, test_name):
        """Test the exceptions of the 'set_parameter' method."""
        with pytest.raises(exception_type):
            ma_def.evaluate(**variables)


# --------------------------------------------------------------------------------------
# TimeSeries
# --------------------------------------------------------------------------------------


class TestTimeSeries:
    """Tests for the TimeSeries class."""

    # Fixtures, aliases and shared variables -------------------------------------------

    ME = MathematicalExpression
    DTI = pd.DatetimeIndex
    TDI = pd.TimedeltaIndex
    TS = TimeSeries

    time_discrete = pd.TimedeltaIndex([0, 1, 2, 3, 4], unit="s")
    value_constant = Q_(1, "m")
    values_discrete = Q_(np.array([10, 11, 12, 14, 16]), "mm")
    me_expr_str = "a*t + b"
    me_params = {"a": Q_(2, "m/s"), "b": Q_(-2, "m")}

    me_params_vec = {"a": Q_([[2, 0, 1]], "m/s"), "b": Q_([[-2, 3, 0]], "m")}

    ts_constant = TimeSeries(value_constant)
    ts_disc_step = TimeSeries(values_discrete, time_discrete, "step")
    ts_disc_linear = TimeSeries(values_discrete, time_discrete, "linear")
    ts_expr = TimeSeries(ME(me_expr_str, me_params))
    ts_expr_vec = TimeSeries(ME(me_expr_str, me_params_vec))

    # test_construction_discrete -------------------------------------------------------

    @staticmethod
    @pytest.mark.parametrize(
        "data, time, interpolation, shape_exp",
        [
            (Q_(1, "m"), None, None, (1,)),
            (Q_([3, 7, 1], "m"), TDI([0, 1, 2], unit="s"), "step", (3,)),
            (Q_([3, 7, 1], ""), Q_([0, 1, 2], "s"), "step", (3,)),
        ],
    )
    def test_construction_discrete(data, time, interpolation, shape_exp):
        """Test the construction of the TimeSeries class."""
        # set expected values
        if isinstance(time, pint.Quantity):
            time_exp = pd.TimedeltaIndex(time.magnitude, unit="s")
        else:
            time_exp = time

        # create instance
        ts = TimeSeries(data=data, time=time, interpolation=interpolation)

        # check
        assert np.all(ts.data == data)
        assert np.all(ts.time == time_exp)
        assert ts.interpolation == interpolation
        assert ts.shape == shape_exp
        assert data.check(UREG.get_dimensionality(ts.units))

        assert np.all(ts.data_array.data == data)
        assert ts.data_array.attrs["interpolation"] == interpolation
        if time_exp is None:
            assert "time" not in ts.data_array
        else:
            assert np.all(ts.data_array.time == time_exp)

    # test_construction_expression -----------------------------------------------------

    params_scalar = {"a": Q_(2, "1/s"), "b": Q_(-2, "")}
    params_vec = {"a": Q_([[2, 3, 4]], "m/s"), "b": Q_([[-2, 3, 1]], "m")}

    @staticmethod
    @pytest.mark.parametrize(
        "data,  shape_exp, unit_exp",
        [
            (ME("a*t + b", params_scalar), (1,), ""),
            (ME("a*t + b", params_vec), (1, 3), "m"),
        ],
    )
    def test_construction_expression(data, shape_exp, unit_exp):
        """Test the construction of the TimeSeries class."""
        ts = TimeSeries(data=data)

        # check
        assert ts.data == data
        assert ts.time is None
        assert ts.interpolation is None
        assert ts.shape == shape_exp
        assert ts.data_array is None
        assert Q_(1, unit_exp).check(UREG.get_dimensionality(ts.units))

    # test_construction_exceptions -----------------------------------------------------

    values_def = Q_([5, 7, 3, 6, 8], "m")
    time_def = Q_([0, 1, 2, 3, 4], "s")
    me_too_many_vars = ME("a*t + b", {})
    me_param_units = ME("a*t + b", {"a": Q_(2, "1/s"), "b": Q_(-2, "m")})
    me_time_vec = ME("a*t + b", {"a": Q_([2, 3, 4], "1/s"), "b": Q_([-2, 3, 1], "")})

    @staticmethod
    @pytest.mark.parametrize(
        "data, time, interpolation, exception_type, test_name",
        [
            (values_def, time_def, "int", ValueError, "# unknown interpolation"),
            (values_def, time_def, None, ValueError, "# wrong interp. parameter type"),
            (values_def, time_def.magnitude, "step", ValueError, "# invalid time type"),
            (me_too_many_vars, None, None, Exception, "# too many free variables"),
            (me_param_units, None, None, Exception, "# incompatible parameter units"),
            (me_time_vec, None, None, Exception, "# not compatible with time vectors"),
            ("a string", None, None, TypeError, "# wrong data type"),
        ],
        ids=get_test_name,
    )
    def test_construction_exceptions(
        data, time, interpolation, exception_type, test_name
    ):
        """Test the exceptions of the 'set_parameter' method."""
        with pytest.raises(exception_type):
            TimeSeries(data=data, time=time, interpolation=interpolation)

    # test_comparison -------------------------------------

    time_wrong_values = TDI([0, 1, 2, 3, 5], unit="s")
    values_discrete_wrong = Q_(np.array([10, 11, 12, 15, 16]), "mm")
    values_unit_wrong = Q_(np.array([10, 11, 12, 14, 16]), "s")
    values_unit_prefix_wrong = Q_(np.array([10, 11, 12, 14, 16]), "m")
    params_wrong_values = {"a": Q_(2, "1/s"), "b": Q_(-1, "")}
    params_wrong_unit = {"a": Q_(2, "g/s"), "b": Q_(-2, "g")}
    params_wrong_unit_prefix = {"a": Q_(2, "m/ms"), "b": Q_(-2, "m")}

    @staticmethod
    @pytest.mark.parametrize(
        "ts, ts_other, result_exp",
        [
            (ts_constant, TS(value_constant), True),
            (ts_disc_step, TS(values_discrete, time_discrete, "step"), True),
            (ts_expr, TS(ME(me_expr_str, me_params)), True),
            (ts_constant, ts_disc_step, False),
            (ts_constant, ts_expr, False),
            (ts_disc_step, ts_expr, False),
            (ts_constant, 1, False),
            (ts_disc_step, 1, False),
            (ts_expr, 1, False),
            (ts_constant, "wrong", False),
            (ts_disc_step, "wrong", False),
            (ts_expr, "wrong", False),
            (ts_constant, TS(Q_(1337, "m")), False),
            (ts_constant, TS(Q_(1, "mm")), False),
            (ts_constant, TS(Q_(1, "s")), False),
            (ts_disc_step, TS(values_discrete, time_wrong_values, "step"), False),
            (ts_disc_step, TS(values_discrete_wrong, time_discrete, "step"), False),
            (ts_disc_step, TS(values_unit_prefix_wrong, time_discrete, "step"), False),
            (ts_disc_step, TS(values_discrete, time_discrete, "linear"), False),
            (ts_expr, TS(ME("a*t + 2*b", me_params)), False),
            (ts_expr, TS(ME(me_expr_str, params_wrong_values)), False),
            (ts_expr, TS(ME(me_expr_str, params_wrong_unit)), False),
            (ts_expr, TS(ME(me_expr_str, params_wrong_unit_prefix)), False),
        ],
    )
    def test_comparison(ts, ts_other, result_exp):
        """Test the TimeSeries comparison methods."""
        assert (ts == ts_other) is result_exp
        assert (ts != ts_other) is not result_exp

    # test_interp_time -----------------------------------------------------------------

    time_single = pd.TimedeltaIndex([2.1], "s")
    time_single_q = Q_(2.1, "s")
    time_mul = pd.TimedeltaIndex([-3, 0.7, 1.1, 1.9, 2.5, 3, 4, 7], "s")
    time_mul_q = Q_([-3, 0.7, 1.1, 1.9, 2.5, 3, 4, 7], "s")
    results_exp_vec = [
        [-8, 3, -3],
        [-0.6, 3, 0.7],
        [0.2, 3, 1.1],
        [1.8, 3, 1.9],
        [3, 3, 2.5],
        [4, 3, 3],
        [6, 3, 4],
        [12, 3, 7],
    ]

    @staticmethod
    @pytest.mark.parametrize(
        "ts, time, magnitude_exp, unit_exp",
        [
            (ts_constant, time_single, 1, "m"),
            (ts_constant, time_single_q, 1, "m"),
            (ts_constant, time_mul, [1, 1, 1, 1, 1, 1, 1, 1], "m"),
            (ts_constant, time_mul_q, [1, 1, 1, 1, 1, 1, 1, 1], "m"),
            (ts_disc_step, time_single, 12, "mm"),
            (ts_disc_step, time_single_q, 12, "mm"),
            (ts_disc_step, time_mul, [10, 10, 11, 11, 12, 14, 16, 16], "mm"),
            (ts_disc_step, time_mul_q, [10, 10, 11, 11, 12, 14, 16, 16], "mm"),
            (ts_disc_linear, time_single, 12.2, "mm"),
            (ts_disc_linear, time_single_q, 12.2, "mm"),
            (ts_disc_linear, time_mul, [10, 10.7, 11.1, 11.9, 13, 14, 16, 16], "mm"),
            (ts_disc_linear, time_mul_q, [10, 10.7, 11.1, 11.9, 13, 14, 16, 16], "mm"),
            (ts_expr, time_single, 2.2, "m"),
            (ts_expr, time_single_q, 2.2, "m"),
            (ts_expr, time_mul, [-8, -0.6, 0.2, 1.8, 3, 4, 6, 12], "m"),
            (ts_expr, time_mul_q, [-8, -0.6, 0.2, 1.8, 3, 4, 6, 12], "m"),
            (ts_expr_vec, time_single, [[2.2, 3, 2.1]], "m"),
            (ts_expr_vec, time_single_q, [[2.2, 3, 2.1]], "m"),
            (ts_expr_vec, time_mul, results_exp_vec, "m"),
        ],
    )
    def test_interp_time(ts, time, magnitude_exp, unit_exp):
        """Test the interp_time function."""
        result = ts.interp_time(time)

        assert np.all(np.isclose(result.data.magnitude, magnitude_exp))
        assert Q_(1, str(result.data.units)) == Q_(1, unit_exp)

        if isinstance(time, pint.Quantity):
            assert np.all(result.time == ut.to_pandas_time_index(time))
        else:
            assert np.all(result.time == time)

    # test_interp_time_exceptions ------------------------------------------------------

    @staticmethod
    @pytest.mark.parametrize("ts", [ts_constant, ts_disc_step, ts_disc_linear, ts_expr])
    @pytest.mark.parametrize(
        "time,  exception_type, test_name",
        [
            (DTI(["2010-10-10"]), ValueError, "# wrong type #1"),
            ("a string", ValueError, "# wrong type #2"),
            ([1, 2, 3], ValueError, "# wrong type #3"),
            (1, ValueError, "# wrong type #4"),
            (Q_(2, "s/m"), Exception, "# wrong type #5"),
        ],
        ids=get_test_name,
    )
    def test_interp_time_exceptions(ts, time, exception_type, test_name):
        """Test the exceptions of the 'set_parameter' method."""
        with pytest.raises(exception_type):
            ts.interp_time(time)



"""Run a complete test on asdf example setup from asdf_template_run.py."""

import asdf

from weldx.asdf.extension import WeldxAsdfExtension, WeldxExtension
from weldx.asdf.tags.weldx.core.iso_groove import get_groove
from weldx.asdf.tags.weldx.custom.testclass import TestClass
from weldx.constants import WELDX_QUANTITY as Q_

obj = TestClass(
    prop1="ASDF",
    prop2=10,
    prop3=10,
    prop4=True,
    list_prop=["str1", "str2"],
    pint_prop=Q_(3, "m"),
    groove_prop=get_groove(
        groove_type="VGroove",
        **dict(t=Q_(8, "mm"), alpha=Q_(60, "deg"), c=Q_(4, "mm"), b=Q_(2, "mm")),
    ),
)

filename = "asdf_template.asdf"
tree = {"obj": obj}

# Write the data to a new file
with asdf.AsdfFile(
    tree,
    extensions=[WeldxExtension(), WeldxAsdfExtension()],
    ignore_version_mismatch=False,
) as ff:
    ff.write_to(filename, all_array_storage="inline")

# read back data from ASDF file
with asdf.open(
    filename, copy_arrays=True, extensions=[WeldxExtension(), WeldxAsdfExtension()]
) as af:
    data = af.tree
    print(data["obj"])



# Licensed under a 3-clause BSD style license - see LICENSE
# -*- coding: utf-8 -*-

from asdf.types import CustomType, ExtensionTypeMeta

__all__ = ["WeldxType", "WeldxAsdfType", "_weldx_types", "_weldx_asdf_types"]

_weldx_types = set()
_weldx_asdf_types = set()


class WeldxTypeMeta(ExtensionTypeMeta):
    """Metaclass to populate _weldx_types and _weldx_asdf_types."""

    def __new__(mcls, name, bases, attrs):
        cls = super().__new__(mcls, name, bases, attrs)

        if cls.organization == "weldx.bam.de" and cls.standard == "weldx":
            _weldx_types.add(cls)
        elif cls.organization == "stsci.edu" and cls.standard == "asdf":
            _weldx_asdf_types.add(cls)

        return cls


class WeldxType(CustomType, metaclass=WeldxTypeMeta):
    """This class represents types that have schemas and tags that are defined
    within weldx.

    """

    organization = "weldx.bam.de"
    standard = "weldx"


class WeldxAsdfType(CustomType, metaclass=WeldxTypeMeta):
    """This class represents types that have schemas that are defined in the ASDF
    standard, but have tags that are implemented within weldx.

    """

    organization = "stsci.edu"
    standard = "asdf"



from dataclasses import dataclass
from typing import List

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

from .gas_component import GasComponent

__all__ = ["ShieldingGasType", "ShieldingGasTypeType"]


@dataclass
class ShieldingGasType:
    """<CLASS DOCSTRING>"""

    gas_component: List[GasComponent]
    common_name: str
    designation: str = None


class ShieldingGasTypeType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/process/shielding_gas_type"
    version = "1.0.0"
    types = [ShieldingGasType]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = ShieldingGasType(**tree)
        return obj



from dataclasses import dataclass

import pint

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.asdf.validators import wx_unit_validator

__all__ = ["UnitValidatorTestClass", "UnitValidatorTestClassType"]


@dataclass
class UnitValidatorTestClass:
    """Testclass for validating pint.Quantities with wx_unit."""

    length_prop: pint.Quantity
    velocity_prop: pint.Quantity
    current_prop: pint.Quantity
    nested_prop: dict
    simple_prop: dict


class UnitValidatorTestClassType(WeldxType):
    """Serialization testclass custom validators."""

    name = "debug/test_unit_validator"
    version = "1.0.0"
    types = [UnitValidatorTestClass]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {
        "wx_unit": wx_unit_validator,
    }

    @classmethod
    def to_tree(cls, node: UnitValidatorTestClass, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = UnitValidatorTestClass(**tree)
        return obj



from dataclasses import dataclass

import pint

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

__all__ = ["BaseMetal", "BaseMetalType"]


@dataclass
class BaseMetal:
    """<CLASS DOCSTRING>"""

    common_name: str
    product_form: str
    thickness: pint.Quantity
    m_number: str = None
    group_number: str = None
    diameter: float = None
    specification_number: str = None
    specification_version: str = None
    specification_organization: str = None
    UNS_number: str = None
    CAS_number: str = None
    heat_lot_identification: str = None
    composition: str = None
    manufacturing_history: str = None
    service_history: str = None
    applied_coating_specification: str = None


class BaseMetalType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/design/base_metal"
    version = "1.0.0"
    types = [BaseMetal]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: BaseMetal, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = BaseMetal(**tree)
        return obj



"""Test all ASDF groove implementations."""

import matplotlib.pyplot as plt
import pytest

from weldx.asdf.tags.weldx.core.iso_groove import (
    BaseGroove,
    _create_test_grooves,
    get_groove,
)
from weldx.asdf.utils import _write_read_buffer
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.geometry import Profile

test_params = _create_test_grooves()


@pytest.mark.parametrize(
    "groove, expected_dtype", test_params.values(), ids=test_params.keys()
)
def test_asdf_groove(groove: BaseGroove, expected_dtype):
    """Test ASDF functionality for all grooves.

    Parameters
    ----------
    groove:
       Groove instance to be tested.

    expected_dtype:
       Expected type of the groove to be tested.

    """
    k = "groove"
    tree = {k: groove}

    data = _write_read_buffer(tree)
    assert isinstance(
        data[k], expected_dtype
    ), f"Did not match expected type {expected_dtype} on item {data[k]}"
    # test content equality using dataclass built-in functions
    assert (
        groove == data[k]
    ), f"Could not correctly reconstruct groove of type {type(groove)}"
    # test to_profile
    assert isinstance(
        groove.to_profile(), Profile
    ), f"Error calling plot function of {type(groove)} "

    # call plot function
    fig, ax = plt.subplots()
    groove.plot(ax=ax)
    plt.close(fig)


def test_asdf_groove_exceptions():
    """Test special cases and exceptions of groove classes."""
    # test parameter string generation
    v_groove = get_groove(
        groove_type="VGroove",
        workpiece_thickness=Q_(9, "mm"),
        groove_angle=Q_(50, "deg"),
        root_face=Q_(4, "mm"),
        root_gap=Q_(2, "mm"),
    )

    assert set(v_groove.param_strings()) == {
        "alpha=50 deg",
        "b=2 mm",
        "c=4 mm",
        "t=9 mm",
    }

    # test custom groove axis labels
    fig, _ = plt.subplots()
    v_groove.plot(axis_label=["x", "y"])
    plt.close(fig)

    # test exceptions
    with pytest.raises(ValueError):
        get_groove(
            groove_type="WrongGrooveString",
            workpiece_thickness=Q_(9, "mm"),
            groove_angle=Q_(50, "deg"),
        )

    with pytest.raises(NotImplementedError):
        BaseGroove().to_profile()

    with pytest.raises(ValueError):
        get_groove(
            groove_type="FrontalFaceGroove",
            workpiece_thickness=Q_(2, "mm"),
            workpiece_thickness2=Q_(5, "mm"),
            groove_angle=Q_(80, "deg"),
            root_gap=Q_(1, "mm"),
            code_number="6.1.1",
        ).to_profile()



"""Test GMAW welding process schema implementation."""

import pytest

from weldx import Q_
from weldx.asdf.utils import _write_read_buffer
from weldx.core import TimeSeries
from weldx.welding import GmawProcess


@pytest.mark.parametrize(
    "inputs",
    [
        GmawProcess(
            "spray",
            "Fronius",
            "TPSi",
            dict(
                wire_feedrate=Q_(10, "m/min"),
                voltage=TimeSeries(Q_(40.0, "V")),
                impedance=Q_(10.0, "percent"),
                characteristic=Q_(5, "V/A"),
            ),
        ),
        GmawProcess(
            "spray",
            "CLOOS",
            "Quinto",
            dict(
                wire_feedrate=Q_(10, "m/min"),
                voltage=TimeSeries(Q_([40.0, 20.0], "V"), Q_([0.0, 10.0], "s")),
                impedance=Q_(10.0, "percent"),
                characteristic=Q_(5, "V/A"),
            ),
            tag="CLOOS/spray_arc",
        ),
        GmawProcess(
            "pulse",
            "CLOOS",
            "Quinto",
            dict(
                wire_feedrate=Q_(10.0, "m/min"),
                pulse_voltage=Q_(40.0, "V"),
                pulse_duration=Q_(5.0, "ms"),
                pulse_frequency=Q_(100.0, "Hz"),
                base_current=Q_(60.0, "A"),
            ),
            tag="CLOOS/pulse",
            meta={"modulation": "UI"},
        ),
        GmawProcess(
            "pulse",
            "CLOOS",
            "Quinto",
            dict(
                wire_feedrate=Q_(10.0, "m/min"),
                pulse_current=Q_(0.3, "kA"),
                pulse_duration=Q_(5.0, "ms"),
                pulse_frequency=Q_(100.0, "Hz"),
                base_current=Q_(60.0, "A"),
            ),
            tag="CLOOS/pulse",
            meta={"modulation": "II"},
        ),
    ],
)
def test_gmaw_process(inputs):
    data = _write_read_buffer({"root": inputs})
    assert data["root"] == inputs



from io import BytesIO
from pprint import pprint as pp

import asdf
import numpy as np
import pandas as pd

from weldx.asdf.extension import WeldxAsdfExtension, WeldxExtension

# Timedelta -------------------------------------------------------
td = pd.Timedelta("5m3ns")

# Timedelta -------------------------------------------------------
td_max = pd.Timedelta("106751 days 23:47:16.854775")

# TimedeltaIndex -------------------------------------------------------
tdi = pd.timedelta_range(start="-5s", end="25s", freq="3s")
tdi_nofreq = pd.TimedeltaIndex([0, 1e9, 5e9, 3e9])

# Timestamp -------------------------------------------------------
ts = pd.Timestamp("2020-04-15T16:47:00.000000001")
ts_tz = pd.Timestamp("2020-04-15T16:47:00.000000001", tz="Europe/Berlin")

# DatetimeIndex -------------------------------------------------------
dti = pd.date_range(start="2020-01-01", periods=5, freq="1D")
dti_tz = pd.date_range(start="2020-01-01", periods=5, freq="1D", tz="Europe/Berlin")
dti_infer = pd.DatetimeIndex(["2020-01-01", "2020-01-02", "2020-01-03", "2020-01-04"])
dti_nofreq = pd.DatetimeIndex(["2020-01-01", "2020-01-02", "2020-01-04", "2020-01-05"])

# ASDF -------------------------------------------------------
filename = "asdf_times.yaml"
tree = dict(
    td=td,
    td_max=td_max,
    tdi=tdi,
    tdi_nofreq=tdi_nofreq,
    ts=ts,
    ts_tz=ts_tz,
    dti=dti,
    dti_infer=dti_infer,
    dti_nofreq=dti_nofreq,
)

# Write the data to a new file
with asdf.AsdfFile(
    tree,
    extensions=[WeldxExtension(), WeldxAsdfExtension()],
    ignore_version_mismatch=False,
) as ff:
    ff.write_to(filename)

# read back data from ASDF file
with asdf.open(
    filename, copy_arrays=True, extensions=[WeldxExtension(), WeldxAsdfExtension()]
) as af:
    data = af.tree

pp(data)

assert isinstance(data, dict)
for k, v in tree.items():
    assert np.all(data[k] == v)

# Write the data to buffer
with asdf.AsdfFile(
    tree,
    extensions=[WeldxExtension(), WeldxAsdfExtension()],
    ignore_version_mismatch=False,
) as ff:
    buff = BytesIO()
    ff.write_to(buff)
    buff.seek(0)

# read back data from ASDF file contents in buffer
with asdf.open(
    buff, copy_arrays=True, extensions=[WeldxExtension(), WeldxAsdfExtension()]
) as af:
    data = af.tree
assert isinstance(data, dict)



import os
from pathlib import Path

from asdf.util import filepath_to_url

SCHEMA_PATH = str(Path(__file__).resolve().parents[0] / "schemas")
WELDX_SCHEMA_URI_BASE = "http://weldx.bam.de/schemas/"
WELDX_URL_MAPPING = [
    (
        WELDX_SCHEMA_URI_BASE,
        filepath_to_url(os.path.join(SCHEMA_PATH, "weldx.bam.de"))
        + "/{url_suffix}.yaml",
    )
]



from dataclasses import dataclass

import pandas as pd

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.asdf.validators import wx_property_tag_validator

__all__ = ["PropertyTagTestClass", "PropertyTagTestClassType"]


@dataclass
class PropertyTagTestClass:
    """Helper class to test the shape validator"""

    prop1: pd.Timestamp = pd.Timestamp("2020-01-01")
    prop2: pd.Timestamp = pd.Timestamp("2020-01-02")
    prop3: pd.Timestamp = pd.Timestamp("2020-01-03")


class PropertyTagTestClassType(WeldxType):
    """Helper class to test the shape validator"""

    name = "debug/test_property_tag"
    version = "1.0.0"
    types = [PropertyTagTestClass]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {
        "wx_property_tag": wx_property_tag_validator,
    }

    @classmethod
    def to_tree(cls, node: PropertyTagTestClass, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = PropertyTagTestClass(**tree)
        return obj
